#!coding:utf-8
"""
The Models API of HepAI Platform.
Author: Zhengde Zhang
All rights reserved, 2023. 
This script is free for non-commercial use. Please contact zdzhang@ihep.ac.cn for commercial use.
"""

import os, sys
import requests
from pathlib import Path
import damei as dm
here = Path(__file__).parent
try:
    import hai
except:
    sys.path.append(str(here.parent.parent.parent))
    import hai
import base64
from .sam.pre_process import sam_pre_process, sam_post_process
from .nougat.pdf_process import pdf_process

logger = dm.getLogger('model.py')

class PreProcessFunctions:
   
    @staticmethod
    def pre_process(model, data):
        """
        面向切面的编程，输入数据经过预处理函数，再传入模型，若未定义，则不进行预处理。
        """
        if model == "meta/segment_anything_model":
            data = sam_pre_process(data)
        elif model == "hepai/hainougat":
            data = pdf_process(data)
        return data

class PostProcessFunctions:
    @staticmethod
    def post_process(model, data):
        """面向切面的变成，输出结果经过后处理函数，再输出，若未定义，则不进行后处理"""
        if model == "meta/segment_anything_model":
            data = sam_post_process(data)
        return data
    

class HaiModel(object):

    @staticmethod
    def ensure_api_key(api_key):
        if api_key is None:
            api_key = os.environ.get("HEPAI_API_KEY", None)
        if api_key is None:
            raise ValueError(f"""            
The HepAI API-KEY is required. You can set it via `hai.api_key=xxx` in your code, or set the environment variable `HEPAI_API_KEY` via `export HEPAI_API_KEY=xxx`.
Alternatively, it can be provided by passing in the `api_key` parameter when calling the method.
""")
        return api_key
    

    @staticmethod
    def list_models(**kwargs):
        return HaiModel.list(**kwargs)

    @staticmethod
    def list(**kwargs):
        """
        List all models on HepAI Platform.
        :param refresh: Whether to refresh the model list, default is False.
        :param return_all_info: Whether to return all information of the models, default is False.
        :param api_key: Your HepAI api key, can be obtained in https://ai.ihep.ac.cn, only return public models if not provided.
        :return: The list of models.
        """  
        api_key = kwargs.pop("api_key", None) or hai.api_key
        api_key = HaiModel.ensure_api_key(api_key)
        
        url = kwargs.pop("url", None)
        if not url:
            host = kwargs.pop("host", "aiapi.ihep.ac.cn")
            port = kwargs.pop("port", None)  # 42901
            if port is None:
                url = f'https://{host}'  
            else:
                url = f"http://{host}:{port}"
        assert url, f'url or (host and port) is required. For example: url="http://aiapi.ihep.ac.cn:42901"'

        ret = requests.post(
            f"{url}/list_models",
            headers={"Authorization": f"Bearer {api_key}"},
            json=kwargs,
            )
        if ret.status_code != 200:
            raise ValueError(f"Hai Model connect url: {url} Error: \n{ret.status_code} {ret.reason} {ret.text}")
        return ret.json()
    
    @staticmethod
    def all_workers_info(**kwargs):
        models = HaiModel.list()
        api_key = kwargs.pop("api_key", None)
        api_key = HaiModel.ensure_api_key(api_key)
        host = kwargs.get("host", "chat.ihep.ac.cn")
        port = kwargs.get("port", None)
        assert api_key, f'api_key is required.'
        if port:
            url = f"http://{host}:{port}"
        else:
            url = f'https://{host}'
        headers={"Authorization": f"Bearer {api_key}"}
        all_info = dict()
        for model_name in models['models']:
            response = requests.post(
                f"{url}/get_worker_info",
                headers=headers,
                json={
                    "model": model_name,
                },
            )
            worker_info = response.json()['info']
            workers = worker_info[model_name]  # 一个模型可能有多个worker
            workers_info = dict()
            for one_worker in workers:
                address = one_worker.pop('address')
                
                workers_info[address] = one_worker
            # all_info.update(worker_info)
            all_info[model_name] = workers_info
            # print(res.json())
        return all_info
    
    
    @staticmethod
    def inference(model, **kwargs):
        """
        Call the model to execute inference.
        :param model: The model name.
        :param api_key: Your HepAI api key, can be obtained in https://ai.ihep.ac.cn.
        :param timeout：The timeout of the request, default is 60s.
        :param stream: Whether to stream the response, default is False.
        :param kwargs: The input data of the model, they are differ for each model, e.g. {"img": "https://xxx.jpg"}.
        """
        api_key = kwargs.pop("api_key", None)
        stream = kwargs.get("stream", False)
        timeout = kwargs.get("timeout", 60)
        pdfbin = kwargs.get('pdfbin', None)
        api_key = api_key or hai.api_key
        

        url = kwargs.pop("url", None)
        if not url:
            host = kwargs.pop("host", "aiapi.ihep.ac.cn")
            port = kwargs.pop("port", None)
            if port is not None:
                url = f"http://{host}:{port}"
            else:
                url = f"https://{host}"

        data = dict()
        data['model'] = model
        data.update(kwargs)

        assert api_key, """
The HepAI API-KEY is required. Please set the environment variable `HEPAI_API_KEY` via `export HEPAI_API_KEY=xxx`.
Alternatively, it can be provided by passing in the `api_key` parameter when calling the `chat` method.
"""

        if not pdfbin:
            data = PreProcessFunctions.pre_process(model=model, data=data)
        else:
            pdfbin= base64.b64encode(pdfbin).decode()
            data['pdfbin'] = pdfbin

        logger.info(f"Requesting {url} ...")
        # logger.info(f"Requesting data: {data}")
        session = requests.Session()
        response = session.post(
        # response = requests.post(
            f'{url}/v1/inference',
            headers={"Authorization": f"Bearer {api_key}"},
            json=data,
            timeout=timeout,
            stream=stream
        )
        
        if response.status_code != 200:
            raise Exception(f"Got status code {response.status_code} from server: {response.text}")
            # print(response.content)  # 只有非流式相应才能查看内容
        try:
            data = response.json()
            status_code = data.pop('status_code', 42901)
            if status_code != 42901:
                error_info = f'Request error: {data}'
                logger.error(error_info)
                raise Exception(error_info)
        except:
            pass
        
        if not stream:            
            data = response.json()
            data = data['message']
        else: 
            data = response.content.decode('utf-8', errors='ignore')
            data = data.replace('[DONE]', "")
        data = PostProcessFunctions.post_process(model=model, data=data)
        return data
        
    @staticmethod
    def sam(**kwargs):
        """
        Segmeng an image via Segment Anything Model.
        :param img: The image path or image array.
        :param input_points: The input points prompt, e.g. [[x1,y1], [x2,y2], ...], defualt is None.
        :param input_labels: The input labels corresponding to the points prompt, 0 represents the background point, 1 represents the front attraction, and the format is [0, 1,...], which needs to be combined with the input points correspond one by one, if not provided, default to all previous attractions.
        :param input_boxes: The input boxes prompt, e.g. [[x1,y1,x2,y2], [x1,y1,x2,y2], ...], defualt is None.
        :return: The segmentation result. if no prompt is provided, the result is the segmentation of the entire image by front attraction grid points.
        """
        model = "meta/segment_anything_model"
        model2 = kwargs.pop("model", None)
        assert model2 is None or model2 == model, f"model={model2} is not supported, only support model={model}"
        return HaiModel.inference(model=model, **kwargs)
    
    @classmethod
    def dalle3(cls, **kwargs):
        print(cls, kwargs)
        return None
    

if __name__ == '__main__':
    ret = HaiModel.list(
        url="http://aiapi.ihep.ac.cn:42901",
    )
    print(len(ret))
    print(ret)



import os
from pathlib import Path
import argparse
from dataclasses import dataclass
import random
import string
import subprocess
import time
import re


def parse_args():
    parser = argparse.ArgumentParser(description='HepAI ECS command line tool to run virtual machines.')
    parser.add_argument('--gres', type=str, default="vgpu:1", help='Generic resource, default is `vgpu:1`.')
    parser.add_argument('--kvm-name', type=str, default="almalinux9.5", help='Name of the KVM configuration, default is `almalinux9.5`.')
    parser.add_argument('--not-save-changes', action='store_true', help="Not save changes of the vitural machine if True, default is False.")
    parser.add_argument('--partition', type=str, default="vgpu", help="Partition to use, default is `vgpu`.")
    parser.add_argument('--kvm-param', type=str, default="auto", help="KVM parameters, format is `num_cores:num_memory`, default is auto.")
    parser.add_argument('-N', '--nodes', type=int, default=1, help="Number of nodes.")
    parser.add_argument('--chdir', type=str, help="Change to directory before running job.")
    parser.add_argument('--nodelist', type=str, help="Specifies the list of nodes to use.")
    parser.add_argument('--exclude', type=str, help="Specifies the list of nodes to exclude.")
    parser.add_argument('-J', '--job-name', type=str, default="auto", help="Name of the job.")
    parser.add_argument('-t', '--time', default="121m", help="Walltime of the machine, default is 120 minutes.")
    parser.add_argument('-d', '--daemon', action='store_true', help="Running in the background, only submitting task")
    parser.add_argument('-q', '--qos', type=str, default="normal", help="Set Quality of Service")

    args = parser.parse_args()

    return args


@dataclass
class NodeResource:
    num_cores: int = 64  # 64 CPU cores
    num_memory: int = 960*0.534  # 960GB
    num_cards: int = 8  # 8 GPUs


class HaiECS:

    def __init__(self, args):
        self.args = args
        self.nr = NodeResource()
        self.run_dir = f"{Path.home()}/VMSTORE"  # 这个决定slrun-xx.out的输出目录



    def check_kvm_name(self, name: str):
        """检查kvm_name对应的镜像是否存在，如果不存在就报错"""
        images_dir = f"{Path.home()}/IMAGES"
        tips = f"Image `{name}` not found, usually it is being configured, please wait or contact helpdesk.ihep.ac.cn."
        if not os.path.exists(images_dir):
            # raise ValueError(f"IMAGES directory not found in {images_dir}.")
            raise ValueError(tips)
        # 检查后缀为.qcow2的文件
        exist_images = [x[:-6] for x in os.listdir(images_dir) if x.endswith(".qcow2")]
        if name not in exist_images:
            # 自动申请创建镜像
            # raise ValueError(f"KVM image `{name}` not found in {images_dir}.")
            raise ValueError(tips)
        print(f"KVM image `{name}` found in {images_dir}.")

        pass

    def ensure_sleep_sh(self, time: str):
        """创建sleep.sh文件，用于指定多少时间后关闭虚拟机"""
        sh_content = """
#!/bin/bash

JOBID=
disk_id=$(ls -1 /dev/disk/by-id| grep -i "virtio-JOBID-" | head -n1)
if [[ $disk_id =~ virtio-JOBID-([0-9]+)_([0-9]+) ]]; then
        JOBID=${BASH_REMATCH[1]}
        MPI=${BASH_REMATCH[2]}
fi
echo "ECS is ready" > /mnt/VMSTORE/ok.finish.${JOBID}

"""
        sh_content += f"sleep {time}\n"


        home = Path.home()
        file_path = f'{home}/VMSTORE/sleep_{time}.sh'
        if not os.path.exists(file_path):
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            with open(file_path, 'w') as f:
                f.write(sh_content)
            # chmod +x
            os.system(f"chmod +x {file_path}")
        return file_path

    def auto_num_cpu_and_memory(self, num_cards: int):
        """根据卡的数量自动设置cpu核数和memory数量"""
        if num_cards > 8:
            raise ValueError(f"Number of cards should be no more than 8, but got {num_cards}")
        assert isinstance(num_cards, int)

        ratio = num_cards / self.nr.num_cards
        n_cores = int(self.nr.num_cores * ratio)
        n_memory = int(self.nr.num_memory * ratio)
        return n_cores, n_memory
    
    
    def auto_job_name(self):
        """自动生成job_name"""
        job_name = f'ecs_'
        # 后面随机四个符号
        for _ in range(4):
            job_name += random.choice(string.ascii_letters).lower()
        return job_name

    def get_sbatch_vm_cmd(self):
        """从args中获取sbatch-vm的命令"""
        args = self.args
        cmd = "sbatch-vm"
        if args.kvm_name:
            self.check_kvm_name(args.kvm_name)  # 检查kvm镜像是否存在
            cmd += f" --kvm-name={args.kvm_name}"

        if args.not_save_changes:
            pass
        else:
            cmd += f" --kvm-edit"  # 需要保存是，设置--kvm-edit

        if args.gres:
            """转换slurm的gres为vslurm的gres"""
            card_type, num_cards = args.gres.split(":")
            if "gpu" in card_type:
                card_type = "vfiogpu"
            elif "dcu" in card_type:
                card_type = "vfiodcu"
            else:
                raise ValueError(f"Unknown gpu/duc card type: {card_type}")
            cmd += f" --gres={card_type}:{num_cards}"

            # 进一步根据卡的数量自动设置cpu核数和memory数量
            if args.kvm_param == "auto":
                num_cores, num_memory = self.auto_num_cpu_and_memory(int(num_cards))
            else:
                num_cores, num_memory = args.kvm_param.split(":")
            cmd += f" --kvm-param={num_cores}:{num_memory}"


        if args.partition:
            cmd += f" --partition={args.partition}"

        if args.nodes:
            cmd += f" --nodes={args.nodes}"

        if args.chdir:
            cmd += f" --chdir={args.chdir}"

        if args.nodelist:
            cmd += f" --nodelist={args.nodelist}"

        if args.exclude:
            cmd += f" --exclude={args.exclude}"

        if args.job_name:
            if args.job_name == "auto":
                args.job_name = self.auto_job_name()

            cmd += f" --job-name={args.job_name}"

        if args.time:
            sleep_file = self.ensure_sleep_sh(args.time)
            cmd += f" {sleep_file}"

        if args.qos:
            cmd += f" --qos={args.qos}"

        cmd += f" --account=ihepai"

        return cmd
    
    
    def run_command(self, command, verbose=True, ignore_error=False):
        try:
            # 通过 subprocess.run 执行命令并获取输出
            if verbose:
                print(f"Run command: {command}")
            result = subprocess.run(command, shell=True, check=True, text=True, capture_output=True)
            # result.stdout 包含命令的标准输出结果

            stdout = result.stdout

            # 可选：result.stderr 也可以获取标准错误输出
            if result.stderr:
                # print("Command Error:")
                # print(result.stderr)
                raise ValueError(f"An error occurred while executing the command: \n{result.stderr}")
            if verbose:
                print(stdout)
            return stdout

        except subprocess.CalledProcessError as e:
            ##在异常中同时打印 stdout 和 stderr
            if verbose:
                print(f"Failed to excute command: {command}")
                print("Standard Output:")
                print(e.stdout)  # 如果可能，还有成功部分的输出
                print("Standard Error:")
                print(e.stderr)  # 打印错误输出
                if "There is a task with '--kvm-edit' parameter" in str(e.stdout):
                    print(f'You already have another editable ECS running. Please use `scancel <job_id>` to terminate it before attempting to start a new ECS. Alternatively, you can set `--not-save-changes` to disable editable mode and run multiple ECS instances.')
            if ignore_error:  # 忽略错误时，不会退出
                return
            else:
                exit(1)  # 退出脚本以指示失败

    def extract_job_id(self, stdout):
        """
        从命令的输出中提取job_id
        如：Command Output:
currentpath=/aifs/user/home/zdzhang/scripts
userhome=/aifs/user/home/zdzhang
numberOfJob=0, maxNumberOfJob=3
sbatch --gres=vfiogpu:1,vfioeth:1 --kvm-param=8:120 --partition=vgpu --nodes=1 --job-name=ecs_uxra /aifs/user/home/zdzhang/VMSTORE/SBATCH/S26658
Submitted batch job 63
        """
        lines = stdout.split("\n")
        for line in lines:
            if "Submitted batch job" in line:
                job_id = line.split(" ")[-1]
                return job_id
        print("Job ID not found.")
        return None

    def get_hostnames_and_ips(self, job_id):
        showip_rst = self.run_command(f"showip {job_id}", verbose=False, ignore_error=True)

        if not showip_rst:
            return None, None

        # 从如下内容中提取ip
        """
        aigpu001, IP: (0, '10.5.6.122', '10.249.6.122', 'aa:0a:05:06:0b:02', 0)
        aigpu002, IP: (0, '10.5.6.130', '10.249.6.130', 'aa:0a:05:06:0c:00', 1)
        """
        # 仅仅提取出10.5.6.122和10.5.6.130
        # pattern = r"(\w+), IP: \(.*?'(10\.\d+\.\d+\.\d+)"
        pattern = r"(\w+), IP: \((.*?)\)"

        # 匹配结果
        matches = re.findall(pattern, showip_rst)

        # 如果找到了IP地址，即返回第一个匹配的IP地址
        if matches:
            hostnames = []
            ips = []
            for match in matches:
                hostname = match[0]
                _, ip, ip2, mac, id_ = eval(match[1])
                hostnames.append(f'{hostname}-{id_}')
                ips.append(ip)
            # print(f"showip_rst: {showip_rst}")
            # print(f"matches: {matches}")
            # print(f"hostnames: {hostnames}")
            # print(f"ips: {ips}")
            return hostnames, ips
        else:
            return None, None

    def check_ecs_is_ready(self, job_id: str):
        ok_file_path = f'{Path.home()}/VMSTORE/ok.finish.{job_id}'
        if os.path.exists(ok_file_path):
            return True
        return False
    
    def polling_result(self, job_id, t0):
        ips = None
        # ping_result = None

        actual_check_interval = 5  # 实际检查间隔
        last_check_time = time.time() - actual_check_interval - 2  # 上次检查时间
        timeout, noticed = 180, False  # 超时时间，等待时候过长时，提醒用户可以按Ctrl+C退出，并hai-ecs -d重新提交任务到后台
        while True:
            """改成每0.2s打印一次，但实际检查5秒一次"""
            if time.time() - last_check_time > actual_check_interval:
                hostnames, ips = self.get_hostnames_and_ips(job_id)
                # 已经有ip，还需要通过ping来判断是否已经启动
                is_ready = self.check_ecs_is_ready(job_id=job_id)
                # if ping_result:
                if is_ready:
                    print("\n")
                    break
                last_check_time = time.time()
            time.sleep(1)
            show_str = f'\rECS is starting up {time.time()-t0:.2f}s (about 1~3 minutes), ip={ips}.'
            print(show_str, end="")
            if time.time() - t0 > timeout and not noticed:
                print("\n")
                print(f"Notice: ECS creation is taking longer than expected. You can press Ctrl+C to exit and run `hai-ecs -d` to submit the task in the background.")
                noticed = True
        return hostnames, ips

    def __call__(self):
        cdir = os.getcwd()

        sbatch_vm_cmd = self.get_sbatch_vm_cmd()
        # 执行命令
        os.chdir(self.run_dir)
        stdout = self.run_command(sbatch_vm_cmd)
        job_id = self.extract_job_id(stdout)
        if self.args.daemon:
            lines = [f"ECS task is submitted."]
            lines += [f"    job_id: {job_id}."]
            lines += [f"    run `showip {job_id}` to get the ip address."]
            lines += [f"    run `scancel {job_id}` to cancel the job."]
            lines += [f"    run `squeue` to check all jobs."]
            lines += [f"    run `ssh root@<ip>` to connect to the ECS."]
            self.print_lines(lines)
            # print(f"Job {job_id} is submitted.")
            return
        if job_id:
            try:
                t0 = time.time()
                hostnames, ips = self.polling_result(job_id=job_id, t0=t0)
            except KeyboardInterrupt:
                print("\nKeyboardInterrupt: ECS creation is interrupted.")
                #self.run_command(f"scancel {job_id}")
                print(f"You can run `squeue-vm` to show the task information or run `hai-ecs -d` to submit a task in the background.")
                exit(1)

        prt_data = {
            "job_id": job_id,
            "startup_time": f"{time.time()-t0:.2f}s",
            "max_walltime": self.args.time,
            "hostname": hostnames[0] if len(hostnames) == 1 else hostnames,
            "ip": ips[0] if len(ips) == 1 else ips,
        }

        max_key_len = max([len(k) for k in prt_data.keys()])

        lines = [f"ECS is ready! "]
        lines += [f"    {k:<{max_key_len}} : {v}" for k, v in prt_data.items()]
        # lines += [f"You can now connect to it via: `ssh root@{ip}`.",]
        for hostname, ip in zip(hostnames, ips):
            lines += [f"    run `ssh root@{ip}` to access `{hostname}`."]

        self.print_lines(lines)

    def print_lines(self, lines):
        max_len = max([len(line) for line in lines])
        print(f"#--{'-'*max_len}--#")
        for line in lines:
            print(f'#  {line:<{max_len}}  #')
        print(f"#--{'-'*max_len}--#")


if __name__ == '__main__':
    args = parse_args()
    # 执行创建或管理VM的逻辑
    hai_ecs = HaiECS(args)
    hai_ecs()


# from damei.nn.uaii.stream.base_module import AbstractModule
# from damei.nn.uaii.stream.base_input import AbstractInput
# from damei.nn.uaii.stream.base_output import AbstractOutput
# from damei.nn.uaii.stream.base_queue import AbstractQue

from hai.uaii.stream.base_module import AbstractModule
from hai.uaii.stream.base_input import AbstractInput
from hai.uaii.stream.base_output import AbstractOutput
from hai.uaii.stream.base_queue import AbstractQue




# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union
from typing_extensions import Annotated, TypeAlias

from ...._utils import PropertyInfo
from .file_path_delta_annotation import FilePathDeltaAnnotation
from .file_citation_delta_annotation import FileCitationDeltaAnnotation

__all__ = ["AnnotationDelta"]

AnnotationDelta: TypeAlias = Annotated[
    Union[FileCitationDeltaAnnotation, FilePathDeltaAnnotation], PropertyInfo(discriminator="type")
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .runs import (
    Runs,
    AsyncRuns,
    RunsWithRawResponse,
    AsyncRunsWithRawResponse,
    RunsWithStreamingResponse,
    AsyncRunsWithStreamingResponse,
)
from .threads import (
    Threads,
    AsyncThreads,
    ThreadsWithRawResponse,
    AsyncThreadsWithRawResponse,
    ThreadsWithStreamingResponse,
    AsyncThreadsWithStreamingResponse,
)
from .messages import (
    Messages,
    AsyncMessages,
    MessagesWithRawResponse,
    AsyncMessagesWithRawResponse,
    MessagesWithStreamingResponse,
    AsyncMessagesWithStreamingResponse,
)

__all__ = [
    "Runs",
    "AsyncRuns",
    "RunsWithRawResponse",
    "AsyncRunsWithRawResponse",
    "RunsWithStreamingResponse",
    "AsyncRunsWithStreamingResponse",
    "Messages",
    "AsyncMessages",
    "MessagesWithRawResponse",
    "AsyncMessagesWithRawResponse",
    "MessagesWithStreamingResponse",
    "AsyncMessagesWithStreamingResponse",
    "Threads",
    "AsyncThreads",
    "ThreadsWithRawResponse",
    "AsyncThreadsWithRawResponse",
    "ThreadsWithStreamingResponse",
    "AsyncThreadsWithStreamingResponse",
]





from typing import List, Dict, Any
from dataclasses import dataclass, field

@dataclass
class CursorPage:
    """
    Generic cursor-based pagination class
    """
    object: str
    data: List[any]
    first_id: str = None
    last_id: str = None
    has_more: bool = False




import torch
import torch.nn.functional as F
from tqdm import tqdm

from .utils.dice_score import multiclass_dice_coeff, dice_coeff


def evaluate(net, dataloader, device):
    net.eval()
    num_val_batches = len(dataloader)
    dice_score = 0

    # iterate over the validation set
    for batch in tqdm(dataloader, total=num_val_batches, desc='Validation round', unit='batch', leave=False):
        image, mask_true = batch['image'], batch['mask']
        # move images and labels to correct device and type
        image = image.to(device=device, dtype=torch.float32)
        mask_true = mask_true.to(device=device, dtype=torch.long)
        mask_true = F.one_hot(mask_true, net.n_classes).permute(0, 3, 1, 2).float()

        with torch.no_grad():
            # predict the mask
            mask_pred = net(image)

            # convert to one-hot format
            if net.n_classes == 1:
                mask_pred = (F.sigmoid(mask_pred) > 0.5).float()
                # compute the Dice score
                dice_score += dice_coeff(mask_pred, mask_true, reduce_batch_first=False)
            else:
                mask_pred = F.one_hot(mask_pred.argmax(dim=1), net.n_classes).permute(0, 3, 1, 2).float()
                # compute the Dice score, ignoring background
                dice_score += multiclass_dice_coeff(mask_pred[:, 1:, ...], mask_true[:, 1:, ...], reduce_batch_first=False)

           

    net.train()

    # Fixes a potential division by zero error
    if num_val_batches == 0:
        return dice_score
    return dice_score / num_val_batches




import asyncio
import logging

from typing import Any, AsyncGenerator, List, Sequence, Callable

from autogen_core import (
    AgentId,
    CancellationToken,
    SingleThreadedAgentRuntime,
    AgentRuntime
)

from autogen_agentchat.base import ChatAgent, TaskResult, TerminationCondition
from autogen_agentchat.messages import (
    BaseAgentEvent, 
    AgentEvent, 
    BaseChatMessage, 
    ChatMessage, 
    MessageFactory,
    ModelClientStreamingChunkEvent, 
    StopMessage,
    TextMessage)
from autogen_agentchat.teams._group_chat._events import (
    GroupChatStart, 
    GroupChatTermination,
    SerializableException,
    )
from autogen_agentchat.teams._group_chat._sequential_routed_agent import SequentialRoutedAgent
from autogen_agentchat.teams._group_chat._base_group_chat_manager import BaseGroupChatManager
from autogen_agentchat.teams import BaseGroupChat

from hepai.agents.modules.managers.base_thread import Thread
from hepai.agents.modules.managers.threads_manager import ThreadsManager
from hepai.agents.modules.managers.base_thread_message import ThreadMessage, Content, Text

event_logger = logging.getLogger(__name__)



class DrSaiGroupChatManager(BaseGroupChatManager):

    def __init__(
        self,
        name: str,
        group_topic_type: str,
        output_topic_type: str,
        participant_topic_types: List[str],
        participant_names: List[str],
        participant_descriptions: List[str],
        output_message_queue: asyncio.Queue[BaseAgentEvent | BaseChatMessage | GroupChatTermination],
        termination_condition: TerminationCondition | None,
        max_turns: int | None,
        message_factory: MessageFactory,
        emit_team_events: bool = False,
        thread: Thread = None,
        thread_mgr: ThreadsManager = None,
        **kwargs: Any
    ):
        
        super().__init__(
            name=name,
            group_topic_type=group_topic_type,
            output_topic_type=output_topic_type,
            participant_topic_types=participant_topic_types,
            participant_names=participant_names,
            participant_descriptions=participant_descriptions,
            output_message_queue=output_message_queue,
            termination_condition=termination_condition,
            max_turns=max_turns,
            message_factory=message_factory,
            emit_team_events=emit_team_events,
        )
        self._theard: Thread = thread
        self._thread_mgr: ThreadsManager = thread_mgr



class DrSaiGroupChat(BaseGroupChat):

    component_type = "team"

    def __init__(
        self,
        participants: List[ChatAgent],
        group_chat_manager_name: str,
        group_chat_manager_class: type[SequentialRoutedAgent],
        termination_condition: TerminationCondition | None = None,
        max_turns: int | None = None,
        runtime: AgentRuntime | None = None,
        custom_message_types: List[type[BaseAgentEvent | BaseChatMessage]] | None = None,
        emit_team_events: bool = False,
        thread: Thread = None,
        thread_mgr: ThreadsManager = None,
        **kwargs: Any
    ):
        super().__init__(
            participants=participants,
            group_chat_manager_name=group_chat_manager_name,
            group_chat_manager_class=group_chat_manager_class,
            termination_condition=termination_condition,
            max_turns=max_turns,
            runtime=runtime,
            custom_message_types=custom_message_types,
            emit_team_events=emit_team_events,
            )
        self._thread: Thread = thread
        self._thread_mgr: ThreadsManager = thread_mgr

    def _create_group_chat_manager_factory(
       self,
        name: str,
        group_topic_type: str,
        output_topic_type: str,
        participant_topic_types: List[str],
        participant_names: List[str],
        participant_descriptions: List[str],
        output_message_queue: asyncio.Queue[BaseAgentEvent | BaseChatMessage | GroupChatTermination],
        termination_condition: TerminationCondition | None,
        max_turns: int | None,
        message_factory: MessageFactory,
        **kwargs: Any
    ) -> Callable[[], DrSaiGroupChatManager]:
        def _factory() -> DrSaiGroupChatManager:
            return DrSaiGroupChatManager(
                name = name,
                group_topic_type = group_topic_type,
                output_topic_type = output_topic_type,
                participant_topic_types = participant_topic_types,
                participant_names = participant_names,
                participant_descriptions = participant_descriptions,
                output_message_queue = output_message_queue,
                termination_condition = termination_condition,
                max_turns = max_turns,
                message_factory = message_factory,
                thread = self._thread,
                thread_mgr = self._thread_mgr,
                **kwargs, 
            )

        return _factory
    



from typing import (
    Mapping, Union, Literal, Iterable, Optional, Dict, List, cast, Type,
    Any,
)
import httpx
import os, sys
from pathlib import Path
here = Path(__file__).parent

import inspect

## opneai
from openai import OpenAI, resources
from openai import NOT_GIVEN, Timeout, NotGiven
from openai._types import Headers, Query, Body
from openai.types import Completion, ChatModel
from openai.resources import Chat

# from openai.types.chat.chat_completion_message_param import ChatCompletionMessageParam

# adopt to openai v1.63.2
from openai.resources.chat.completions.completions import (
    ChatCompletionMessageParam,
    completion_create_params,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionToolParam,
    ChatCompletion,
    ChatCompletionChunk,
    Stream,
    Completions,
    ChatCompletionStreamOptionsParam,
)

from openai._utils import required_args, maybe_transform
from openai._base_client import (
    ResponseT, APIResponse, FinalRequestOptions, RAW_RESPONSE_HEADER, _T,
    Stream, SSEDecoder, AsyncStream, SSEBytesDecoder,
    make_request_options, LegacyAPIResponse, BaseAPIResponse,
    get_origin, extract_response_type
)

from hai.apis.workers_api.model import HaiModel
from .utils.general import load_image_from_bytes
from .utils.file_object import HaiFile

DEFAULT_MAX_RETRIES = 2
DEFAULT_TIMEOUT = httpx.Timeout(timeout=600.0, connect=5.0)
DEFAULT_LIMITS = httpx.Limits(max_connections=100, max_keepalive_connections=20)

class HaiPostParser:
    def __call__(self, parsed, **kwargs):
        if isinstance(parsed, ChatCompletion):
            """服务端仅仅返回字典时"""
            if parsed.id is None and parsed.choices is None:
                if parsed.model_extra:
                    return parsed.model_extra

        return parsed

class HAPIResponse(APIResponse):

    def get_filename(self, response: httpx.Response) -> str:
        content_disposition = response.headers.get("content-disposition", "")
        filename = content_disposition.split("filename=")[-1]
        return filename.strip('"')
     
    def _parse(self, *, to: type[_T] | None = None) -> Any | _T:
        response = self.http_response
        content_type, *_ = response.headers.get("content-type", "*").split(";")
        if "image/" in content_type:
            filename = self.get_filename(response)
            return HaiFile(type_="image", data=response.content, filename=filename)
        elif "application/pdf" in content_type:
            filename = self.get_filename(response)
            return HaiFile(type_="pdf", data=response.content, filename=filename)
        elif ("text/" in content_type) and ("event-stream" not in content_type):
            # Fix text/event-stream parse error in 202404
            filename = self.get_filename(response)
            return HaiFile(type_="txt", data=response.content, filename=filename)
        # application/json时
        return super()._parse(to=to)


class HaiCompletion(ChatCompletion):
    pass

class HaiCompletions(Completions):

    @required_args(["messages", "model"], ["messages", "model", "stream"])
    def create(
        self,
        *,
        messages: Iterable[ChatCompletionMessageParam],
        model: Union[str, ChatModel],
        frequency_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        function_call: completion_create_params.FunctionCall | NotGiven = NOT_GIVEN,
        functions: Iterable[completion_create_params.Function] | NotGiven = NOT_GIVEN,
        logit_bias: Optional[Dict[str, int]] | NotGiven = NOT_GIVEN,
        logprobs: Optional[bool] | NotGiven = NOT_GIVEN,
        max_tokens: Optional[int] | NotGiven = NOT_GIVEN,
        n: Optional[int] | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        response_format: completion_create_params.ResponseFormat | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        stop: Union[Optional[str], List[str]] | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        stream_options: Optional[ChatCompletionStreamOptionsParam] | NotGiven = NOT_GIVEN,
        temperature: Optional[float] | NotGiven = NOT_GIVEN,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        tools: Iterable[ChatCompletionToolParam] | NotGiven = NOT_GIVEN,
        top_logprobs: Optional[int] | NotGiven = NOT_GIVEN,
        top_p: Optional[float] | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> ChatCompletion | Stream[ChatCompletionChunk]:
        body=maybe_transform(
                {
                    "messages": messages,
                    "model": model,
                    "frequency_penalty": frequency_penalty,
                    "function_call": function_call,
                    "functions": functions,
                    "logit_bias": logit_bias,
                    "logprobs": logprobs,
                    "max_tokens": max_tokens,
                    "n": n,
                    "presence_penalty": presence_penalty,
                    "response_format": response_format,
                    "seed": seed,
                    "stop": stop,
                    "stream": stream,
                    "stream_options": stream_options,
                    "temperature": temperature,
                    "tool_choice": tool_choice,
                    "tools": tools,
                    "top_logprobs": top_logprobs,
                    "top_p": top_p,
                    "user": user,
                },
                completion_create_params.CompletionCreateParams,
            )
        options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            )
        res = self._post(
            "/chat/completions", 
            body=body, options=options,
            cast_to=ChatCompletion,
            stream=stream or False,
            stream_cls=Stream[ChatCompletionChunk],
        )

        return res


    def request_worker(
            self,
            *,
            model: str,
            function: str,
            stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
            extra_headers: Headers | None = None,
            extra_query: Query | None = None,
            extra_body: Body | None = None,
            timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
            **kwargs,
        ):
        tmp = dict()
        tmp["model"] = model
        tmp["function"] = function
        tmp.update(kwargs)
        body=maybe_transform(tmp, completion_create_params.CompletionCreateParams)
        options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            )
        options["post_parser"] = HaiPostParser()
        res = self._post(
            "/worker/unified_gate", 
            body=body, 
            options=options,
            cast_to=HaiCompletion,
            stream=stream or False,
            stream_cls=Stream[ChatCompletionChunk],
        )
        return res

    def verify_api_key(
            self, 
            api_key: str, 
            extra_headers: Headers | None = None,
            extra_query: Query | None = None,
            extra_body: Body | None = None,
            timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
            **kwargs):
        
        params = {
            "api_key": api_key,
        }
        params.update(kwargs)
        options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            )
        try:
            res = self._post(
                "/verify_api_key", 
                body=params, 
                options=options,
                cast_to=HaiCompletion,
                stream=False,
            )
            rst = res.model_extra
        except Exception as e:
            status_code = getattr(e, "status_code", None)
            if status_code == 401:  # 验证失败
            # if status_code.status_code == 401:  # 验证失败
                response = getattr(e, "response", None)
                detail = response.json().get("detail", "")
                rst = {"success": False, "detail": detail}
            else:
                raise e
        return rst



class HaiChat(Chat):
    @property
    def completions(self) -> HaiCompletions:
        return HaiCompletions(client=self._client)

class HepAI(OpenAI):
    completions: HaiCompletions
    chat: HaiChat

    def __init__(
        self,
        *,
        api_key: str | None = None,
        organization: str | None = None,
        base_url: str | httpx.URL | None = None,
        timeout: Union[float, Timeout, None, NotGiven] = NOT_GIVEN,
        max_retries: int = DEFAULT_MAX_RETRIES,
        default_headers: Mapping[str, str] | None = None,
        default_query: Mapping[str, object] | None = None,
        http_client: httpx.Client | None = None,

        _strict_response_validation: bool = False,
        proxy: str | None = None,
    ) -> None:
        if (http_client is None) and (proxy is not None):
            http_client = self.get_http_client(proxy, base_url=base_url, timeout=timeout,)

        super().__init__(
            api_key=api_key,
            organization=organization,
            base_url=base_url,
            timeout=timeout,
            max_retries=max_retries,
            default_headers=default_headers,
            default_query=default_query,
            http_client=http_client,
            _strict_response_validation=_strict_response_validation,
        )
        
        self.completions = HaiCompletions(client=self)
        self.chat = HaiChat(client=self)

        ## 2024.11.13 更新支持DDF Client
        from hepai.modules.haiddf.client.haiddf_client import HaiDDFClient
        timeout = None if (timeout == NOT_GIVEN or timeout is None) else timeout
        self.ddf = HaiDDFClient(api_key=api_key, base_url=base_url, timeout=timeout, proxy=proxy)
    
    @classmethod
    def get_http_client(cls, proxy, **kwargs) -> httpx.Client:
        if proxy is None:
            return None
        else:
            proxies = {
                "http://": proxy,
                "https://": proxy,
            }
        base_url = kwargs.get("base_url", None)
        base_url = base_url or "https://aiapi.ihep.ac.cn/v1"
        timeout = kwargs.get("timeout", DEFAULT_TIMEOUT)
        timeout = DEFAULT_TIMEOUT if (timeout == NOT_GIVEN or timeout is None) else timeout
        transport = kwargs.get("transport", None)
        limits = kwargs.get("limits", DEFAULT_LIMITS)
        limits = DEFAULT_LIMITS if (limits == NOT_GIVEN or limits is None) else limits
        http_client = httpx.Client(
            base_url=base_url,
            timeout=timeout,
            proxies=proxies,
            transport=transport,
            limits=limits,
        )
        return http_client

    def list_models(self, **kwargs):
        api_key = kwargs.pop("api_key", self.api_key)
        host = kwargs.pop("host", self.base_url.host)
        port = kwargs.pop("port", self.base_url.port)
        return HaiModel.list(
            api_key=api_key,
            host=host,
            port=port,
            **kwargs,
        )
    
    def verify_api_key(self, api_key: str = None, **kwargs):
        api_key = api_key or self.api_key
        return self.completions.verify_api_key(api_key, **kwargs)
    
    def request_worker(self, **kwargs):
        return self.completions.request_worker(**kwargs)

    def _process_response(
        self,
        *,
        cast_to: Type[ResponseT],
        options: FinalRequestOptions,
        response: httpx.Response,
        stream: bool,
        stream_cls: type[Stream[Any]] | type[AsyncStream[Any]] | None,
        retries_taken: int = 0,
    ) -> ResponseT:
        if response.request.headers.get(RAW_RESPONSE_HEADER) == "true":
            return cast(
                ResponseT,
                LegacyAPIResponse(
                    raw=response,
                    client=self,
                    cast_to=cast_to,
                    stream=stream,
                    stream_cls=stream_cls,
                    options=options,
                    retries_taken=retries_taken,
                ),
            )

        origin = get_origin(cast_to) or cast_to

        if inspect.isclass(origin) and issubclass(origin, BaseAPIResponse):
            if not issubclass(origin, APIResponse):
                raise TypeError(f"API Response types must subclass {APIResponse}; Received {origin}")

            response_cls = cast("type[BaseAPIResponse[Any]]", cast_to)
            return cast(
                ResponseT,
                response_cls(
                    raw=response,
                    client=self,
                    cast_to=extract_response_type(response_cls),
                    stream=stream,
                    stream_cls=stream_cls,
                    options=options,
                    retries_taken=retries_taken,
                ),
            )

        if cast_to == httpx.Response:
            return cast(ResponseT, response)

        api_response = HAPIResponse(
            raw=response,
            client=self,
            cast_to=cast("type[ResponseT]", cast_to),  # pyright: ignore[reportUnnecessaryCast]
            stream=stream,
            stream_cls=stream_cls,
            options=options,
            retries_taken=retries_taken,
        )
        if bool(response.request.headers.get(RAW_RESPONSE_HEADER)):
            return cast(ResponseT, api_response)

        return api_response.parse()
    

    ### --- 关于DDF Client的函数 --- ###
    ## # --- 关于User的函数 --- ##
    def list_users(self):
        """
        List all users
        Note: Only admin can use this function
        """
        return self.ddf.user.list_users()
    
    def create_user(self, **kwargs):
        """
        Create a new user
        Note: Only admin can use this function
        """
        return self.ddf.user.create_user(**kwargs)
    
    def delete_user(self, user_id: str):
        """
        Delete a user
        Note: Only admin can use this function
        """
        return self.ddf.user.delete_user(user_id=user_id)
    
    ### --- 关于Key的函数 --- ###
    def list_api_keys(self):
        """
        List all keys
        Note: Only admin can use this function
        """
        return self.ddf.key.list_api_keys()
    
    def create_api_key(self, **kwargs):
        """
        Create a new key
        Note: Only admin can use this function
        """
        return self.ddf.key.create_api_key(**kwargs)
    
    def delete_api_key(self, api_key_id: str):
        """
        Delete a key
        Note: Only admin can use this function
        """
        return self.ddf.key.delete_api_key(api_key_id=api_key_id)
    
    def get_remote_model(self, model_name: str):
        """
        Get a remote model
        """
        return self.ddf.worker.get_remote_model(model_name=model_name)


"""
用来管理不同的用户所创建的助手的类
"""
import os, sys
from pathlib import Path
here = Path(__file__).parent
import time

try:
    from hepai.agents.version import __version__
except:
    sys.path.insert(1, f'{here.parent.parent.parent}')
    from hepai.agents.version import __version__
from hepai.agents.utils import BaseJsonSaver, Logger
from hepai.agents.version import __appname__

from .base_assistant import Assistant, Tool, AssistantDeleted

logger = Logger.get_logger("assistants_manager.py")

class AssistantsManager(BaseJsonSaver):
    """
    离线存储所有的AI助手
    """

    version = "1.0.0"
    metadata = {
        "description": "AI assistants manager of all users",
        "mapping_username2indexes": {},  # 用来存储用户名到线程索引的映射，方便快速查找
    }
    
    def __init__(self,
        file_name: str = f'assistants.json',
        file_dir: str = f'{Path.home()}/.{__appname__}',
        **kwargs
        ) -> None:
        super().__init__(auto_save=True, **kwargs)

        self.file_path = os.path.join(file_dir, file_name)

        self._data = self._init_load(self.file_path, version=self.version, metadata=self.metadata)
        pass


    def get_assistant(self, assistant_id, username=None, return_index=False):
        """获取助手"""
        if username:
            assts_indexes = self.data["metadata"]["mapping_username2indexes"].get(username, [])
            if not assts_indexes:
                raise ModuleNotFoundError(f"No assistant found for user `{username}`")
        else:
            assts_indexes = list(range(len(self.entities)))
            if not assts_indexes:
                raise ModuleNotFoundError(f"No assistant found")
        filtered = [(self.entities[idx], idx) for idx in assts_indexes if self.entities[idx]['id'] == assistant_id]
        if not filtered:
            raise ModuleNotFoundError(f"No assistant found for user `{username}` and id `{assistant_id}`")
        assts, indexes = zip(*filtered)
        if not assts:
            raise ModuleNotFoundError(f"Assistant not found, id: `{assistant_id}`")
        if len(assts) > 1:
            raise ValueError(f"Multiple assistants found, id: `{assistant_id}`")
        asst = Assistant(**assts[0])
        if asst.deleted:
            raise ModuleNotFoundError(f"Assistant has been deleted, id: `{assistant_id}`")
        if return_index:
            assert len(indexes) == 1, f"Multiple assistants found, id: `{assistant_id}`"
            return asst, indexes[0]
        return asst


    def create_assistant(self, model, username=None, **kwargs):
        """
        返回值：
        ```json
        {
            "id": "asst_gbPJLxkrpsawUo9opT7SZh2O",
            "object": "assistant",
            "created_at": 1714452287,
            "name": "Math Tutor",
            "description": null,
            "model": "gpt-4-turbo",
            "instructions": "You are a personal math tutor. Write and run code to answer math questions.",
            "tools": [{
                "type": "code_interpreter"
                }],
            "top_p": 1.0,
            "temperature": 1.0,
            "file_ids": [],
            "metadata": {},
            "response_format": "auto"
            }'
        """
        username = username or self.DEFAULT_USERNAME
        id = self.auto_id(prefix='asst_', length=30)

        # order_id = f'{len(self.entities)+1:0>6}'  # 这个不对
        order_ids = [x['order_id'] for x in self.entities]
        order_id = f'{int(max(order_ids))+1:0>6}' if order_ids else '000001'
        
        asst = Assistant(
            id=id,
            object="assistant",
            created_at=int(time.time()),
            name=kwargs.get('name', None),
            description=kwargs.get('description', None),
            model=model,
            instructions=kwargs.get('instructions', None),
            tools=kwargs.get('tools', []),
            tool_resources=kwargs.get('tool_resources', {}),
            top_p=kwargs.get('top_p', 1),
            temperature=kwargs.get('temperature', 0.6),
            # file_ids=kwargs.get('file_ids', []),
            metadata=kwargs.get('metadata', {}),
            response_format=kwargs.get('response_format', 'auto'),

            username=username,
            order_id=order_id
        )
        self.append_entity(asst, username=username, save_immediately=False)

        # 保存的和输出的数据可能不一样
        return asst
        # return self.fit_output(obj, **kwargs)

    def delete_assistant(self, assistant_id, username=None, **kwargs):
        """删除助手"""
        save_immediately = kwargs.pop('save_immediately', False)
        permanent = kwargs.pop('permanent', False)  # 是否永久删除
        permanent = True if permanent == 'true' else permanent  # 适配restful api
        if not permanent:  # 软删除，添加deleted标记
            asst = self.update_assistant(
                assistant_id, username=username, deleted=True,
                save_immediately=save_immediately
                )
            assert asst.deleted, "Assistant not deleted"
        else:
            asst, idx = self.get_assistant(assistant_id, username=username, return_index=True)
            self.remove_entity(asst, idx, save_immediately=save_immediately)

        return AssistantDeleted(
            id=asst.id,
            object="assistant.deleted",
            deleted=True,
        )
    
    def update_assistant(self, assistant_id, username=None, **kwargs):
        """更新助手"""
        save_immediately = kwargs.pop('save_immediately', False)
        asst, idx = self.get_assistant(assistant_id, username=username, return_index=True)
        asst.update(**kwargs)  # 更新类属性

        # 更新到数据库
        self.update_entity(asst, idx, save_immediately=save_immediately)

        # logger.debug(f"asst: {asst}")
        return asst

    def retrieve_assistant(self, assistant_id, username=None, **kwargs):
        """获取模型"""
        return self.get_assistant(assistant_id, username=username, **kwargs)
    
    def list_assistants(self, username, limit=20, order='desc', after=None, before=None):
        """
        list assistants of a user
        {
        "object": "list",
        "data": [
            {
            "id": "asst_abc123",
            "object": "assistant",
            "created_at": 1698982736,
            "name": "Coding Tutor",
            "description": null,
            "model": "gpt-4-turbo",
            "instructions": "You are a helpful assistant designed to make me better at coding!",
            "tools": [],
            "tool_resources": {},
            "metadata": {},
            "top_p": 1.0,
            "temperature": 1.0,
            "response_format": "auto"
            },
            ...
            ],
        "first_id": "asst_abc123",
        "last_id": "asst_abc789",
        "has_more": false
        }
        """
        if after:
            raise NotImplementedError(f"`after` param is not supported yet in version {__version__}")
        if before:
            raise NotImplementedError(f"`before` param is not supported yet in version {__version__}")

        assts_indexes = self.data["metadata"]["mapping_username2indexes"].get(username, [])
        assts = [self.entities[idx] for idx in assts_indexes]
        assts = [Assistant(**x) for x in assts]
        assts = [x for x in assts if not x.deleted]
        reverse = True if order == 'desc' else False
        # assistants = sorted(assistants, key=lambda x: x['created_at'], reverse=reverse)
        assts = sorted(assts, key=lambda x: x.created_at, reverse=reverse)
        has_more = len(assts) > limit
        if has_more:
            assts = assts[:limit]
        if len(assts) == 0:
            # raise ModuleNotFoundError(f"No assistant found for user `{username}`")
            return {
                "object": "list",
                "data": assts,
                "first_id": None,
                "last_id": None,
                "has_more": False,
            }

        return {
            "object": "list",
            "data": assts,
            "first_id": assts[0].id,
            "last_id": assts[-1].id,
            "has_more": has_more,
        }
        

if __name__ == "__main__":
    am = AssistantsManager(debug=True)

    # 增加
    my_asst = am.create_assistant(
        name='Dr.Sai-Primary', username='anonymous', model='gpt-4-turbo',
        save_immediately=True
        )
    print(f'my_asst: {my_asst}')

    # list
    listed_assts = am.list_assistants(
        username='anonymous', limit=20, order='desc')
    print(f'listed_assts: {listed_assts}')

    # 更新
    up_asst = am.update_assistant(
        assistant_id=my_asst.id, 
        username='anonymous', name='Dr.Sai-Primary-Updated')
    print(f'up_asst: {up_asst}')

    # 获取
    my_asst2 = am.retrieve_assistant(assistant_id=my_asst.id)
    print(f'my_asst2: {my_asst2}')

    # 删除
    del_asst = am.delete_assistant(assistant_id=my_asst.id, permanent=True)
    print(f'del_asst: {del_asst}')




import uuid
import time

from typing import List
from schemas import OpenAIChatMessage

import inspect
from typing import get_type_hints, Literal, Tuple


def stream_message_template(model: str, message: str):
    return {
        "id": f"{model}-{str(uuid.uuid4())}",
        "object": "chat.completion.chunk",
        "created": int(time.time()),
        "model": model,
        "choices": [
            {
                "index": 0,
                "delta": {"content": message},
                "logprobs": None,
                "finish_reason": None,
            }
        ],
    }


def get_last_user_message(messages: List[dict]) -> str:
    for message in reversed(messages):
        if message["role"] == "user":
            if isinstance(message["content"], list):
                for item in message["content"]:
                    if item["type"] == "text":
                        return item["text"]
            return message["content"]
    return None


def get_last_assistant_message(messages: List[dict]) -> str:
    for message in reversed(messages):
        if message["role"] == "assistant":
            if isinstance(message["content"], list):
                for item in message["content"]:
                    if item["type"] == "text":
                        return item["text"]
            return message["content"]
    return None


def get_system_message(messages: List[dict]) -> dict:
    for message in messages:
        if message["role"] == "system":
            return message
    return None


def remove_system_message(messages: List[dict]) -> List[dict]:
    return [message for message in messages if message["role"] != "system"]


def pop_system_message(messages: List[dict]) -> Tuple[dict, List[dict]]:
    return get_system_message(messages), remove_system_message(messages)


def add_or_update_system_message(content: str, messages: List[dict]):
    """
    Adds a new system message at the beginning of the messages list
    or updates the existing system message at the beginning.

    :param msg: The message to be added or appended.
    :param messages: The list of message dictionaries.
    :return: The updated list of message dictionaries.
    """

    if messages and messages[0].get("role") == "system":
        messages[0]["content"] += f"{content}\n{messages[0]['content']}"
    else:
        # Insert at the beginning
        messages.insert(0, {"role": "system", "content": content})

    return messages


def doc_to_dict(docstring):
    lines = docstring.split("\n")
    description = lines[1].strip()
    param_dict = {}

    for line in lines:
        if ":param" in line:
            line = line.replace(":param", "").strip()
            param, desc = line.split(":", 1)
            param_dict[param.strip()] = desc.strip()
    ret_dict = {"description": description, "params": param_dict}
    return ret_dict


def get_tools_specs(tools) -> List[dict]:
    function_list = [
        {"name": func, "function": getattr(tools, func)}
        for func in dir(tools)
        if callable(getattr(tools, func)) and not func.startswith("__")
    ]

    specs = []

    for function_item in function_list:
        function_name = function_item["name"]
        function = function_item["function"]

        function_doc = doc_to_dict(function.__doc__ or function_name)
        specs.append(
            {
                "name": function_name,
                # TODO: multi-line desc?
                "description": function_doc.get("description", function_name),
                "parameters": {
                    "type": "object",
                    "properties": {
                        param_name: {
                            "type": param_annotation.__name__.lower(),
                            **(
                                {
                                    "enum": (
                                        param_annotation.__args__
                                        if hasattr(param_annotation, "__args__")
                                        else None
                                    )
                                }
                                if hasattr(param_annotation, "__args__")
                                else {}
                            ),
                            "description": function_doc.get("params", {}).get(
                                param_name, param_name
                            ),
                        }
                        for param_name, param_annotation in get_type_hints(
                            function
                        ).items()
                        if param_name != "return"
                    },
                    "required": [
                        name
                        for name, param in inspect.signature(
                            function
                        ).parameters.items()
                        if param.default is param.empty
                    ],
                },
            }
        )

    return specs



from hepai.agents import DrSai
import os, sys
from pathlib import Path
here = Path(__file__).parent
from typing import Generator, Optional, Union, Dict
from fastapi import FastAPI, Request, Header
from fastapi.responses import StreamingResponse, HTMLResponse, FileResponse, JSONResponse, PlainTextResponse
from dataclasses import dataclass, field
import json
from fastapi import FastAPI, Form, File, UploadFile, HTTPException, Response
from fastapi import APIRouter, Query
import traceback
import uvicorn
import inspect  
from collections.abc import AsyncGenerator 

try:
    from hepai.agents.version import __version__
except:
    sys.path.append(str(here.parent.parent))
    from hepai.agents.version import __version__

import hepai
from hepai import HepAI, HRModel, HModelConfig, HWorkerConfig, HWorkerAPP

from hepai.agents.utils import Logger

logger = Logger.get_logger("app_worker.py")

class DrSaiAPP(DrSai):
    '''
    chat/completion:路由,直接处理聊天和自动回复请求。
    该路由接收前端页面的请求，并返回相应的回复。
    该路由的请求参数包括:
    - messages: 输入的消息列表

    OpenAI Assistants格式的标准接口的DrSai后端服务:
    该后端服务通过http请求或者hepai的openai assistants格式的标准接口api, 提供Dr.Sai多智能体后端服务。
    包括:
    1. Assistants-相关接口，包括创建、获取、删除、更新助手。用于对接前后端Agents设置和前端页面的交互。
    2. Threads-相关接口，包括创建、获取、删除、更新会话。用于对接前端页面的会话交互。
    3. Runs-相关接口，包括创建、获取、删除、更新运行。用于对接前端页面的运行交互。

    '''
    app = FastAPI()
    router = APIRouter(prefix="/apiv2", tags=["agent"])
    
    def __init__(self, **kwargs):

        super(DrSaiAPP, self).__init__(**kwargs)

        self._init_router()
    
    def _init_router(self):

        # 测试路由
        DrSaiAPP.router.get("/")(self.index)

        # chat/completion路由
        DrSaiAPP.router.post("/chat/completions")(self.a_chat_completions)


    #### --- 关于DrSai的路由 --- ####
    async def index(self, request: Request):
        return f"Hello, world! This is DrSai WebUI {__version__}"
    
    ### --- 关于chat_completions的路由 --- ###
    async def a_chat_completions(self, request: Request):

        headers = request.headers
        if not isinstance(headers, dict):
            headers = dict(headers)
        authorization = headers.get("authorization", None)
        if authorization:
            apikey = authorization.split(" ")[-1]
        else:
            apikey = None

        # apikey = headers.get("authorization").split(" ")[-1]
        params = await request.json()
        params.update({"apikey": apikey})
        if "messages" not in params or "model" not in params:
            raise HTTPException(status_code=400, detail="messages and model must be required, see https://platform.openai.com/docs/api-reference/chat")
        # return self.try_except_raise_http_exception(
        #     self.a_start_chat_completions, **params
        #     )
        return await self.try_except_raise_http_exception(
            self.a_start_chat_completions, **params
            )

    ### --- 其它函数 --- #### 
    async def try_except_raise_http_exception(self, func, *args, **kwargs):  # 改为异步函数
        """智能捕获函数内部raise的异常，转换为HTTPException返回，支持同步/异步函数"""
        try:
            # 判断是否是异步函数
            if inspect.iscoroutinefunction(func):
                res = await func(*args, **kwargs)  # 异步函数需要await
            else:
                res = func(*args, **kwargs)  # 同步函数直接调用

            # 处理同步/异步生成器
            if isinstance(res, (AsyncGenerator, Generator)):  # 同时检查两种生成器
                return StreamingResponse(res)
            return res
    
        except Exception as e:
            tb_str = traceback.format_exception(*sys.exc_info())
            tb_str = "".join(tb_str)
            logger.debug(f"Error: {e}.\nTraceback: {tb_str}")

            e_class = e.__class__.__name__
            error_mapping = {
                "ModuleNotFoundError": 404,
                "NotImplementedError": 501,
                # 添加更多映射...
            }
            status_code = error_mapping.get(e_class, 400)
            raise HTTPException(
                status_code=status_code,
                detail=f'{e_class}("{str(e)}")'
            )
    # def try_except_raise_http_exception(self, func, *args, **kwargs):
    #     """智能捕获函数内部raise的异常，转换为HTTPException返回，便于前端处理"""
    #     try:
    #         res = func(*args, **kwargs)
    #         if isinstance(res, Generator):
    #             return StreamingResponse(res)
    #         return res
    #     except Exception as e:
    #         # 获取报错类型：e.__class__.__name__
    #         # if self.debug:
    #             # logger.error(f"Error: {e}")
    #         # tb_str = traceback.format_exception(etype=type(e), value=e, tb=e.__traceback__)
    #         tb_str = traceback.format_exception(*sys.exc_info())
    #         tb_str = "".join(tb_str)
    #         logger.debug(f"Error: {e}.\nTraceback: {tb_str}")

    #         e_class = e.__class__.__name__
    #         if e_class == "ModuleNotFoundError":
    #             raise HTTPException(status_code=404, detail=f'{e_class}("{str(e)}")')
    #         elif e_class == "NotImplementedError":
    #             raise HTTPException(status_code=501, detail=f'{e_class}("{str(e)}")')
    #         ## TODO: 其他报错类型转换为合适的报错状态码
    #         raise HTTPException(status_code=400, detail=f'{e_class}("{str(e)}")')


from typing import List, Union, Generic, Optional
from typing_extensions import Literal

from ._types import ParsedChatCompletionSnapshot
from ...._models import BaseModel, GenericModel
from ..._parsing import ResponseFormatT
from ....types.chat import ChatCompletionChunk, ChatCompletionTokenLogprob


class ChunkEvent(BaseModel):
    type: Literal["chunk"]

    chunk: ChatCompletionChunk

    snapshot: ParsedChatCompletionSnapshot


class ContentDeltaEvent(BaseModel):
    """This event is yielded for every chunk with `choice.delta.content` data."""

    type: Literal["content.delta"]

    delta: str

    snapshot: str

    parsed: Optional[object] = None


class ContentDoneEvent(GenericModel, Generic[ResponseFormatT]):
    type: Literal["content.done"]

    content: str

    parsed: Optional[ResponseFormatT] = None


class RefusalDeltaEvent(BaseModel):
    type: Literal["refusal.delta"]

    delta: str

    snapshot: str


class RefusalDoneEvent(BaseModel):
    type: Literal["refusal.done"]

    refusal: str


class FunctionToolCallArgumentsDeltaEvent(BaseModel):
    type: Literal["tool_calls.function.arguments.delta"]

    name: str

    index: int

    arguments: str
    """Accumulated raw JSON string"""

    parsed_arguments: object
    """The parsed arguments so far"""

    arguments_delta: str
    """The JSON string delta"""


class FunctionToolCallArgumentsDoneEvent(BaseModel):
    type: Literal["tool_calls.function.arguments.done"]

    name: str

    index: int

    arguments: str
    """Accumulated raw JSON string"""

    parsed_arguments: object
    """The parsed arguments"""


class LogprobsContentDeltaEvent(BaseModel):
    type: Literal["logprobs.content.delta"]

    content: List[ChatCompletionTokenLogprob]

    snapshot: List[ChatCompletionTokenLogprob]


class LogprobsContentDoneEvent(BaseModel):
    type: Literal["logprobs.content.done"]

    content: List[ChatCompletionTokenLogprob]


class LogprobsRefusalDeltaEvent(BaseModel):
    type: Literal["logprobs.refusal.delta"]

    refusal: List[ChatCompletionTokenLogprob]

    snapshot: List[ChatCompletionTokenLogprob]


class LogprobsRefusalDoneEvent(BaseModel):
    type: Literal["logprobs.refusal.done"]

    refusal: List[ChatCompletionTokenLogprob]


ChatCompletionStreamEvent = Union[
    ChunkEvent,
    ContentDeltaEvent,
    ContentDoneEvent[ResponseFormatT],
    RefusalDeltaEvent,
    RefusalDoneEvent,
    FunctionToolCallArgumentsDeltaEvent,
    FunctionToolCallArgumentsDoneEvent,
    LogprobsContentDeltaEvent,
    LogprobsContentDoneEvent,
    LogprobsRefusalDeltaEvent,
    LogprobsRefusalDoneEvent,
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["ImageModel"]

ImageModel: TypeAlias = Literal["dall-e-2", "dall-e-3"]



"""
Here are constant values that are used in the project
"""

import os, sys
from pathlib import Path
here = Path(__file__).parent

from ..version import __appname__, __version__, __author__

## Basic Info
APPNAME = __appname__
AUTHOR = __author__
VERSION = __version__


## Paths
try:
    USERNAME = os.getlogin()  # 当前用户名
except:
    USERNAME = os.getenv('USER') or os.getenv('LOGNAME') or os.getenv('USERNAME')  # 适配WSL和windows环境

REPO_ROOT = f'{here.parent.parent}'  # 项目根目录
PROMPTS_DIR = f'{REPO_ROOT}/DrSai/configs/prompts'  # prompts目录
FS_DIR = f'{Path.home()}/.{APPNAME}'  # 文件系统目录

RUNS_DIR = f'{REPO_ROOT}/runs'  # 运行目录
FILE_DIR = f'{FS_DIR}/files'   # 文件目录
# CONFIG_DIR = f'{FS_DIR}/configs'  # 配置目录
CONFIG_DIR = f'{REPO_ROOT}/DrSai/configs'  # 配置目录

##
DEFAULT_USERNAME = "anonymous"  # for 创建assistant和获取assistant

## logger
LOGGER_DIR = f'{Path.home()}/.{APPNAME}/logs'  # 日志目录
LOGGER_LEVEL = "INFO"  # 日志级别
# LOGGER_LEVEL = "DEBUG"
# LOGGER_LEVEL = "WARNING"

## event
EVENT_TIMEOUT = 60  # 事件等待超时时间
EVENT_INTERVAL = 0.05  # 事件返回间隔时间



from autogen_agentchat.teams import SelectorGroupChat
from autogen_agentchat.teams._group_chat._selector_group_chat import SelectorFuncType, CandidateFuncType
from autogen_agentchat.messages import BaseAgentEvent, BaseChatMessage
from autogen_agentchat.base import ChatAgent, TerminationCondition
from autogen_core.models  import ChatCompletionClient
from autogen_core import AgentRuntime
from hepai.agents import HepAIChatCompletionClient
from hepai.agents.modules.managers.base_thread import Thread
from hepai.agents.modules.managers.threads_manager import ThreadsManager

from typing import List, Optional, Any
import os

class DrSaiSelectorGroupChat(SelectorGroupChat):
    def __init__(
        self,
        participants: List[ChatAgent],
        *,
         model_client: ChatCompletionClient = None,
        termination_condition: TerminationCondition | None = None,
        max_turns: int | None = None,
        runtime: AgentRuntime | None = None,
        selector_prompt: str = """You are in a role play game. The following roles are available:
{roles}.
Read the following conversation. Then select the next role from {participants} to play. Only return the role.

{history}

Read the above conversation. Then select the next role from {participants} to play. Only return the role.
""",
        allow_repeated_speaker: bool = False,
        max_selector_attempts: int = 3,
        selector_func: Optional[SelectorFuncType] = None,
        candidate_func: Optional[CandidateFuncType] = None,
        custom_message_types: List[type[BaseAgentEvent | BaseChatMessage]] | None = None,
        emit_team_events: bool = False,
        model_client_streaming: bool = False,
        thread: Thread = None,
        thread_mgr: ThreadsManager = None,
        **kwargs: Any
    ):
        if not model_client:
            model_client = HepAIChatCompletionClient(model="openai/gpt-4o", api_key=os.environ.get("HEPAI_API_KEY"))
        super().__init__(
            participants=participants,
            model_client=model_client,
            termination_condition=termination_condition, 
            max_turns=max_turns,
            runtime=runtime,
            selector_prompt=selector_prompt,
            allow_repeated_speaker=allow_repeated_speaker,
            max_selector_attempts=max_selector_attempts,
            selector_func=selector_func,
            candidate_func=candidate_func,
            custom_message_types=custom_message_types,
            emit_team_events=emit_team_events,
            model_client_streaming=model_client_streaming,
        )

        self._theard: Thread = thread
        self._thread_mgr: ThreadsManager = thread_mgr




# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # 
# This script is used to collect all events for streaming #
# Author: Zhengde Zhang                                   #
# Date: 2024-05-23                                        #
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

from typing import List, Callable, Generator
import asyncio
import queue
import time
import threading

class EventCollector:
    """
    支持运行时的事件收集
    Usage:
        evc = EventCollector()
        evc.add_event_source(Callable)
        stream = evc.run()
        for x in stream:
            print(f"Event: {x}")
    """
    def __init__(self):
        self.que = queue.Queue()
        self.event_sources: List[Callable] = []

    def add_event_source(self, event_source: Generator):
        """添加一个事件源，该事件源应该是一个生成器"""
        self.event_sources.append(event_source)
    
    def producer(self, gen, q):
        for item in gen:  # 生成器
            q.put(item)
        q.put(None)  # 自动的停止标志

    def consumer(self, q: queue.Queue, max_producers):
        count = 0  # source完成的标志
        while True:
            item = q.get()
            if item is None:
                count += 1
                if count == max_producers:
                    break
            else:
                yield item
            q.task_done()

    def all_events_generator(self):
        """返回流式所有的事件源信息，事件源停止后，不再返回事件"""
        evss = self.event_sources
        
        # res = await asyncio.gather(*evss)
        producers = [
            threading.Thread(target=self.producer, args=(gen, self.que))
            for gen in evss
        ]
        # consumer_thread = threading.Thread(target=self.consumer, args=(self.que, len(producers)))
        consumer = self.consumer(self.que, len(producers))
        
        for p in producers:
            p.start()
        
        # consumer_thread.start()
        return consumer
    
        # 等待所有的生产者结束
        for p in producers:
            p.join()
        # 等待消费者结束
        # consumer_thread.join()

class FakeEventSource_1():
    # async def __anext__(self):
    def event_generator(self):
        # print("FakeEventSource_1")
        # await asyncio.sleep(1)
        for i in range(2):
            time.sleep(1)
            yield {
                "data": "data",
                "event": f"fe1_data_{i}",
            }

class FakeEventSource_2():
    # async def __anext__(self):
    def event_generator(self):
        # print("FakeEventSource_2")
        # await asyncio.sleep(1)
        for i in range(2):
            time.sleep(2)
            yield {
                "data": "data",
                "event": f"fe2_data_{i}",
            }

class FakeEventSource_3():

    def __init__(self) -> None:
        self.fe3_msg = "in1"
    # async def __anext__(self):

    
    def run_chat(self, evc: EventCollector):
        msg = ""
        msg += self.fe3_msg
        for i in range(3):
            round = i+1
            gen = self.gen_reply(msg, round)
            evc.add_event_source(gen)

        pass

    def gen_reply(self, msg, round):
        # print("FakeEventSource_3")
        # await asyncio.sleep(1)
        outputs = [f'{msg}-r{round}-c1', f'{msg}-r{round}-c2']
        full_response = ""
        for i, x in enumerate(outputs):
            time.sleep(1)
            full_response += x
            self.fe3_msg = x
            yield {
                "data": x,
                "event": f"fe3.round{round}.chunk{i}",
            }
        msg += full_response
        
    async def a_run_chat(self, evc: EventCollector):
        msg = "in1"
        for i in range(2):
            round = i+1
            res = await self.gen_reply(msg, round)
            pass


def test():
    evc = EventCollector()
    # evc.add_event_source(FakeEventSource_1().event_generator())
    # evc.add_event_source(FakeEventSource_2().event_generator())
    # stream = asyncio.run(evc.all_events_generator())
    fe3 = FakeEventSource_3()
    # fe3.run_chat(evc=evc)
    fe3.a_run_chat(evc=evc)

    stream = evc.all_events_generator()
    for x in stream:
        print(f"Event: {x}")

async def test():
    env = EventCollector()
    fe3 = FakeEventSource_3()
    await fe3.a_run_chat(env)

    return env.all_events_generator()
        
                
if __name__ == '__main__':
    # test()
    import asyncio
    stream = asyncio.run(test())
    for x in stream:
        print(f"Event: {x}")



from typing import List, Optional
from pydantic import BaseModel
from schemas import OpenAIChatMessage
import os
import requests
import json

from utils.pipelines.main import (
    get_last_user_message,
    add_or_update_system_message,
    get_tools_specs,
)


class Pipeline:
    class Valves(BaseModel):
        # List target pipeline ids (models) that this filter will be connected to.
        # If you want to connect this filter to all pipelines, you can set pipelines to ["*"]
        pipelines: List[str] = []

        # Assign a priority level to the filter pipeline.
        # The priority level determines the order in which the filter pipelines are executed.
        # The lower the number, the higher the priority.
        priority: int = 0

        # Valves for function calling
        OPENAI_API_BASE_URL: str
        OPENAI_API_KEY: str
        TASK_MODEL: str
        TEMPLATE: str

    def __init__(self):
        # Pipeline filters are only compatible with Open WebUI
        # You can think of filter pipeline as a middleware that can be used to edit the form data before it is sent to the OpenAI API.
        self.type = "filter"

        # Optionally, you can set the id and name of the pipeline.
        # Best practice is to not specify the id so that it can be automatically inferred from the filename, so that users can install multiple versions of the same pipeline.
        # The identifier must be unique across all pipelines.
        # The identifier must be an alphanumeric string that can include underscores or hyphens. It cannot contain spaces, special characters, slashes, or backslashes.
        # self.id = "function_calling_blueprint"
        self.name = "Function Calling Blueprint"

        # Initialize valves
        self.valves = self.Valves(
            **{
                "pipelines": ["*"],  # Connect to all pipelines
                "OPENAI_API_BASE_URL": os.getenv(
                    "OPENAI_API_BASE_URL", "https://api.openai.com/v1"
                ),
                "OPENAI_API_KEY": os.getenv("OPENAI_API_KEY", "YOUR_OPENAI_API_KEY"),
                "TASK_MODEL": os.getenv("TASK_MODEL", "gpt-3.5-turbo"),
                "TEMPLATE": """Use the following context as your learned knowledge, inside <context></context> XML tags.
<context>
    {{CONTEXT}}
</context>

When answer to user:
- If you don't know, just say that you don't know.
- If you don't know when you are not sure, ask for clarification.
Avoid mentioning that you obtained the information from the context.
And answer according to the language of the user's question.""",
            }
        )

    async def on_startup(self):
        # This function is called when the server is started.
        print(f"on_startup:{__name__}")
        pass

    async def on_shutdown(self):
        # This function is called when the server is stopped.
        print(f"on_shutdown:{__name__}")
        pass

    async def inlet(self, body: dict, user: Optional[dict] = None) -> dict:
        # If title generation is requested, skip the function calling filter
        if body.get("title", False):
            return body

        print(f"pipe:{__name__}")
        print(user)

        # Get the last user message
        user_message = get_last_user_message(body["messages"])

        # Get the tools specs
        tools_specs = get_tools_specs(self.tools)

        # System prompt for function calling
        fc_system_prompt = (
            f"Tools: {json.dumps(tools_specs, indent=2)}"
            + """
If a function tool doesn't match the query, return an empty string. Else, pick a function tool, fill in the parameters from the function tool's schema, and return it in the format { "name": \"functionName\", "parameters": { "key": "value" } }. Only pick a function if the user asks.  Only return the object. Do not return any other text."
"""
        )

        r = None
        try:
            # Call the OpenAI API to get the function response
            r = requests.post(
                url=f"{self.valves.OPENAI_API_BASE_URL}/chat/completions",
                json={
                    "model": self.valves.TASK_MODEL,
                    "messages": [
                        {
                            "role": "system",
                            "content": fc_system_prompt,
                        },
                        {
                            "role": "user",
                            "content": "History:\n"
                            + "\n".join(
                                [
                                    f"{message['role']}: {message['content']}"
                                    for message in body["messages"][::-1][:4]
                                ]
                            )
                            + f"Query: {user_message}",
                        },
                    ],
                    # TODO: dynamically add response_format?
                    # "response_format": {"type": "json_object"},
                },
                headers={
                    "Authorization": f"Bearer {self.valves.OPENAI_API_KEY}",
                    "Content-Type": "application/json",
                },
                stream=False,
            )
            r.raise_for_status()

            response = r.json()
            content = response["choices"][0]["message"]["content"]

            # Parse the function response
            if content != "":
                result = json.loads(content)
                print(result)

                # Call the function
                if "name" in result:
                    function = getattr(self.tools, result["name"])
                    function_result = None
                    try:
                        function_result = function(**result["parameters"])
                    except Exception as e:
                        print(e)

                    # Add the function result to the system prompt
                    if function_result:
                        system_prompt = self.valves.TEMPLATE.replace(
                            "{{CONTEXT}}", function_result
                        )

                        print(system_prompt)
                        messages = add_or_update_system_message(
                            system_prompt, body["messages"]
                        )

                        # Return the updated messages
                        return {**body, "messages": messages}

        except Exception as e:
            print(f"Error: {e}")

            if r:
                try:
                    print(r.json())
                except:
                    pass

        return body



from hepai.agents.dr_sai import DrSai

# Agent components
from hepai.agents.modules.components.LLMClient import HepAIChatCompletionClient

# Agents
from hepai.agents.modules.baseagent.drsaiagent import DrSaiAgent as AssistantAgent

# Groupchat
from hepai.agents.modules.groupchat._round_robin_group_chat import DrSaiRoundRobinGroupChat, DrSaiRoundRobinGroupChatManager
from hepai.agents.modules.groupchat._selector_group_chat import DrSaiSelectorGroupChat
from hepai.agents.modules.groupchat._swarm_group_chat import DrSaiSwarm
from hepai.agents.modules.groupchat._base_group_chat import DrSaiGroupChatManager, DrSaiGroupChat

# manager
from hepai.agents.modules.managers.base_thread import Thread
from hepai.agents.modules.managers.threads_manager import ThreadsManager
from hepai.agents.modules.managers.base_thread_message import ThreadMessage, Content, Text

# reply functions
from hepai.agents.modules.baseagent.toolagent import tools_reply_function, tools_recycle_reply_function

# # tools
# from hepai.agents.modules.components.tools.mcps_std import web_fetch

# utils
from hepai.agents.utils.message_convert import (
    llm_messages2oai_messages, 
    llm_messages2basechatmessages)

from hepai.agents.utils.oai_stream_event import (
    chatcompletionchunk, 
    chatcompletionchunkend, 
    chatcompletions)

# backend
from hepai.agents.backend.run import (
    Run_DrSaiAPP, 
    run_backend, 
    run_console, 
    run_hepai_worker,
    run_openwebui,
    run_pipelines,
    run_drsai_app)
from hepai.agents.backend.app_worker import DrSaiAPP

###########
# Autogen #
###########

# autogen_ext Models
from autogen_ext.models.openai import OpenAIChatCompletionClient
from autogen_core.models  import (
    ChatCompletionClient,
    ModelCapabilities,  # type: ignore
    ModelFamily,
    ModelInfo,
    validate_model_info,
)

# autogen_agentchat Agents
from autogen_agentchat.agents import (
    UserProxyAgent, 
    BaseChatAgent,
    CodeExecutorAgent, 
    SocietyOfMindAgent)

# autogen_agentchat Groupchat
from autogen_agentchat.teams import (
    BaseGroupChat, 
    RoundRobinGroupChat, 
    Swarm,
    SelectorGroupChat, 
    MagenticOneGroupChat)

# autogen_agentchat Groupchat Termination Conditions
from autogen_agentchat.conditions import (
    ExternalTermination,
    HandoffTermination,
    MaxMessageTermination,
    SourceMatchTermination,
    StopMessageTermination,
    TextMentionTermination,
    TimeoutTermination,
    TokenUsageTermination,)

# autogen_agentchat UI
from autogen_agentchat.ui import Console, UserInputManager

# autogen_agentchat Messages
from autogen_core.models import (
    AssistantMessage,
    ChatCompletionTokenLogprob,
    CreateResult,
    FinishReasons,
    FunctionExecutionResult,
    FunctionExecutionResultMessage,
    LLMMessage,
    RequestUsage,
    SystemMessage,
    TopLogprob,
    UserMessage,
)
from autogen_agentchat.base import Response
from autogen_agentchat.messages import (
    BaseAgentEvent,
    BaseChatMessage,
    ToolCallSummaryMessage,
    ModelClientStreamingChunkEvent,
    TextMessage,
    HandoffMessage,
)

# autogen_core Tools
from autogen_core.tools import (
    Tool, 
    ToolSchema, 
    ParametersSchema,
    BaseTool,
    BaseToolWithState,
    FunctionTool,
    StaticWorkbench,
    ImageResultContent, 
    TextResultContent, 
    ToolResult, 
    Workbench
    )

# autogen_ext mcp
from autogen_ext.tools.mcp import (
    McpServerParams, 
    SseServerParams, 
    StdioServerParams,
    StdioMcpToolAdapter,
    SseMcpToolAdapter,
    McpWorkbench,
    create_mcp_server_session,
    mcp_server_tools)

from autogen_core import Image as AGImage
from autogen_core import CancellationToken



from typing import (
    List,
    Dict, 
    )

from autogen_core.models import (
    FunctionExecutionResultMessage,
    LLMMessage,
    UserMessage,
    AssistantMessage,
    SystemMessage,
)

from autogen_agentchat.messages import (
    # BaseAgentEvent,
    BaseChatMessage,
    ToolCallSummaryMessage,
    # ModelClientStreamingChunkEvent,
    TextMessage,
)

async def llm_messages2oai_messages(llm_messages: List[LLMMessage]) -> List[Dict[str, str]]:
    """Convert a list of LLM messages to a list of OAI chat messages."""
    messages = []
    for llm_message in llm_messages:
        if isinstance(llm_message, SystemMessage):
            messages.append({"role": "system", "content": llm_message.content} )
        if isinstance(llm_message, UserMessage):
            messages.append({"role": "user", "content": llm_message.content, "name": llm_message.source})
        if isinstance(llm_message, AssistantMessage):
            messages.append({"role": "assistant", "content": llm_message.content, "name": llm_message.source})
        if isinstance(llm_message, FunctionExecutionResultMessage):
            messages.append({"role": "function", "content": llm_message.content})
    return messages

async def llm_messages2basechatmessages(
    llm_messages: List[LLMMessage]
    ) -> List[BaseChatMessage]:
    basechatmessages = []
    for llm_message in llm_messages:
        if isinstance(llm_message, SystemMessage):
            continue
        if isinstance(llm_message, FunctionExecutionResultMessage):
            basechatmessages.append(ToolCallSummaryMessage(content=llm_message.content, source="assistant"))
        basechatmessages.append(TextMessage(content=llm_message.content, source=llm_message.source))
    
    return basechatmessages




from dataclasses import dataclass, field
from typing import List, Dict, Any, Optional, Literal, Generator
import time
import json
import uuid
from hepai.agents.configs import CONST

@dataclass
class BaseObject:

    def __post_init__(self):
        self._queue = None  # 用于收集事件的队列

    @property
    def queue(self):
        if self._queue is None:
            import queue
            self._queue = queue.Queue()
        return self._queue

    @property
    def output_keys(self) -> List[str]:
        raise NotImplementedError("Please implement the property `output_keys` in the subclass of `BaseObject`")
    
    @property
    def allowed_update_keys(self) -> List[str]:
        raise NotImplementedError("Please implement the property `allowed_update_keys` in the subclass of `BaseObject`")

    def to_dict(self, only_output_keys=True):
        if only_output_keys:
            return {k: v for k, v in self.__dict__.items() if k in self.output_keys}
        return self.__dict__
    
    def update(self, auto_metadata=False, **new_info):
        """
        更新对象的信息
        :param auto_metadata: 是否自动更新metadata，自动更新"modified"和"modified_at"字段
        """
        for k, v in new_info.items():
            assert k in self.allowed_update_keys, f"Key `{k}` is not allowed to update"
            setattr(self, k, v)
        if auto_metadata:
            self.metadata["modified"] = True
            self.metadata["modified_at"] = int(time.time())
        return self
    
    def get(self, key, default=None):
        return getattr(self, key, default)
    
    def event_generator(self) -> Generator:
        """
        事件生成器，用于流式输出的事件
        """
        end_status = ["completed", "failed", "incomplete", "cancelled", "expired"]
        interval = CONST.EVENT_INTERVAL
        timeout = CONST.EVENT_TIMEOUT
        ctime = time.time()
        while True:
            ### 有事件时，直接返回
            if not self.queue.empty():
                data = self._queue.get()
                # 如果队列中本来就是一个生成器，则返回生成器内容
                if isinstance(data, Generator):
                    for x in data:
                        yield x
                else:  # 否则，直接返回数据
                    yield data
                time.sleep(0.05)
                continue
            # 没有事件，看是否已经结束
            if self.status in end_status:
                break
            # 等待事件
            waited_time = time.time() - ctime
            if waited_time > timeout:
                break
            time.sleep(interval)  

    @property
    def valid_status(self):
        return ["queued", "created", "requires_action", "in_progress", 
                "completed", "failed", "cancelling", "cancelled", "expired"]
    
    def set_status(self, value: str, **kwargs):
        """
        带有事件收集器的状态设置方法
        :param emit: 是否发送事件，事件会被放入队列中并通过事件生成器通过流式输出反馈给前端
        """
        value = value.lower()
        assert value in self.valid_status, f"Invalid status of `ThreadRun`: {value}"
        self.status = value

        if value == "completed":
            self.completed_at = int(time.time())

        emit = kwargs.get("emit", False)
        if emit:
            data = self.construct_status_event()
            sse_data = f"data: {json.dumps(data)}\n\n"
            self.queue.put(sse_data)
            # return self.status_event()
        # return self.status
            
    def auto_id(self, prefix: str = '', length: int = 30, deduplicate: bool | List = True):
        """
        自动生成一个20位的id。
        """
        new_id = uuid.uuid4().hex
        short_id = str(new_id).replace('-', '')
        short_id = prefix + short_id[:length-len(prefix)]
        return short_id
    
    def construct_status_event(self):
        raise NotImplementedError("Please implement the method `construct_status_event` in the subclass of `BaseObject`")
    
    def status_event(self, status=None, set_status=False):
        status = status or self.status
        if set_status:
            if status is None:
                raise ValueError("Status is required when `set_status` is True")
            self.set_status(status, emit=False)
        yield f'data: {json.dumps(self.construct_status_event(status=status))}\n\n'

        # raise NotImplementedError("Please implement the method `status_event` in the subclass of `BaseObject`")
    

@dataclass
class Tool:
    type: str

@dataclass
class Assistant(BaseObject):
    id: str
    object: str
    model: str
    created_at: int
    name: str
    description: Optional[str]
    instructions: str
    tools: List[Tool]
    tool_resources: Dict[str, Any]
    metadata: Dict[str, Any]
    top_p: float
    temperature: float
    response_format: str

    username: str
    order_id: str

    agent_type: Literal["AssistantAgent", "HostAgent", "Human", "Planner", "Coder", "Tester"] = "AssistantAgent"
    deleted: Optional[bool] = False

    @property
    def output_keys(self):
        """定义了网络请求时需要返回的字段"""
        return ["id", "object", "model", "created_at", "name", "description", 
                "instructions", "tools", "tool_resources", "metadata", "top_p",
                "temperature", "response_format", "username", "order_id", "agent_type", "deleted"]
    
    @property
    def allowed_update_keys(self):
        return ["model", "name", "description", "instructions", "tools", "tool_resources", "metadata", "temperature", "top_p", "response_format", "deleted"]
    

@dataclass
class AssistantDeleted:
    id: str
    object: str = "assistant.deleted"
    deleted: bool = True


from openai.types.chat.chat_completion import ChatCompletion, Choice
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk, ChoiceDelta
from openai.types.chat.chat_completion_chunk import Choice as ChunkChoice



# 标准openai chat/completions stream模式
chatcompletionchunk = {
                "id":"chatcmpl-123",
                "object":"chat.completion.chunk",
                "created":1694268190,
                "model":"DrSai", 
                "system_fingerprint": "fp_44709d6fcb", 
                "usage": None,
                "choices":[{"index":0,
                            "delta":{"content":"", "function_call": None, "role": None, "tool_calls": None},
                            "logprobs":None,
                            "finish_reason":None}] # 非None或者stop字段会触发前端askuser
                } 
chatcompletionchunkend = {
    "id":"chatcmpl-123",
    "object":"chat.completion.chunk",
    "created":1694268190,
    "model":"DrSai", 
    "system_fingerprint": "fp_44709d6fcb", 
    "choices":[{'delta': {'content': None, 'function_call': None, 'refusal': None, 'role': None, 'tool_calls': None}, 'finish_reason': 'stop', 'index': 0, 'logprobs': None}]
    }

chatcompletions = {
    'id': 'chatcmpl-123', 
    'choices': [
    {'finish_reason': 'stop', 
    'index': 0, 
    'logprobs': None, 
    'message': {
        'content': '', 
        'refusal': None, 
        'role': 'assistant', 
        'audio': None, 
        'function_call': None, 
        'tool_calls': None}}], 
        'created': 1739758379, 
        'model': 'DrSai', 
        'object': 'chat.completion', 
        'service_tier': 'default', 
        'system_fingerprint': 'fp_13eed4fce1', 
        'usage': {
            'completion_tokens': 10, 'prompt_tokens': 8, 
            'total_tokens': 18, 'completion_tokens_details': {'accepted_prediction_tokens': 0, 'audio_tokens': 0, 'reasoning_tokens': 0, 'rejected_prediction_tokens': 0}, 
            'prompt_tokens_details': {'audio_tokens': 0, 'cached_tokens': 0}}}

def split_string(s, n):
    return [s[i:i+n] for i in range(0, len(s), n)]






from dataclasses import dataclass, field
from typing import Dict, Any, List, Callable
import threading
import time

from .base_assistant import BaseObject
from .base_thread_message import ThreadMessage
from .base_run import ThreadRun

@dataclass
class ToolResources:
    type: str
    url: str
    description: str

@dataclass
class ThreadDeleted:
    id: str
    object: str = "thread.deleted"
    deleted: bool = True


@dataclass
class Thread(BaseObject):
    id: Any  # ID的类型可能取决于具体情况，这里使用Any以保持通用性
    object: str
    created_at: int
    metadata: Dict[str, Any]
    tool_resources: Dict[str, Any]
    
    username: str
    order_id: str
    messages: List[ThreadMessage] = field(default_factory=list)
    runs: List[ThreadRun] = field(default_factory=list)

    deleted: bool = False

    def __post_init__(self):
        self._locked = False  # 标识线程锁定，锁定时无法向Thread添加新的消息，无法在Thread上执行新的Run
        self._running_run = None  # 当前正在运行的Run

        self.thread = None

    @property
    def allowed_update_keys(self):
        return ["metadata", "tool_resources", "messages", "deleted"]
    
    @property
    def output_keys(self):
        # 没username字段
        return ["id", "object", "created_at", "metadata", "tool_resources", 
                "username", "order_id", "messages", "runs", "deleted"]
    
    @property
    def locked(self):
        return self._locked
    

    def start(self, run_func: Callable, run: ThreadRun, **kwargs):
        """
        在Thread上开始一个Run
        """
        assert not self.locked, f"Thread `{self.id}` is locked, cannot start a new run at this time"
        daemon = kwargs.get("daemon", True)

        run.status = "in_progress"
        # 创建一个自线程
        self._running_run = run
        self.thread = threading.Thread(
            # target=self._run, 
            target=run_func,
            args=(run,), 
            daemon=daemon)
        self.thread.start()

        pass

    def _run(self, run: ThreadRun):
        """
        真正执行Run的函数
        """
        count = 0
        while True:
            print(f"Run `{run.id}` is running on thread `{self.id}`... {count}")
            count += 1
            time.sleep(1)
            if count >= 10:
                break
        run.status = "completed"
        run.completed_at = int(time.time())
        # self.stop(run=run)
        pass

    def stop(self, run: ThreadRun = None):
        """
        停止Thread上的所有Run
        """
        run = run or self._running_run
        assert run is not None, "No running run on this thread"
        if run.status == "requires_action":
            run.status = "expired"
            run.expires_at = int(time.time())
        elif run.status == "cancelling":
            run.status = "cancelled"
            run.cancelled_at = int(time.time())
        elif run.status == "completed":
            run.status = "completed"
            run.completed_at = int(time.time())
        elif run.status == "in_progress":
            raise Exception(
                "Cannot stop a running run when in progress, \
                specify a reason to stop it to determine the `failed` or `incomplete` status")
        self.thread.join()

        self._running_run = None
        pass

    



  


import os

from .configs import CONST
from .apis import __version__, __author__, __appname__, __email__, __affiliation__, __version_suffix__
from .version import __url__
from .apis import AbstractInput, AbstractModule, AbstractOutput, AbstractQue
from .apis import MODULES, SCRIPTS, IOS, init_register
from .apis import Config
from .apis import UAII, uaii, cli
from .apis import hub
# from .apis import hai_config as config  # inclue hai root_path, weights_root and other configs
from .apis import grpc_secure_server
from .apis import Testor
from .apis import argparse, parse_args_into_dataclasses, parse_args
from .apis import general
from .apis import worker, BaseWorkerModel, WorkerArgs

# LLM
from .apis import LLM, Model, Models, api_key
# from .apis import HepAI
from .uaii.hepai_object import (
    HepAI, HaiCompletions, ChatCompletion, ChatCompletionChunk, Stream
    )
from .uaii.utils.file_object import HaiFile
from .configs import CONST


# from xsensing_ai.modules import *  # 加载项目的模块
# from xsensing_ai.uaii.server.grpc.grpc_xai_client import XAIGrpcClient

# from . import nn

# """
# 导入内部模块和外部模块
internal_modules = [
    'hai/apis/modules/YOLOv5',
    'hai/apis/modules/ResNet',
    # 'hai/apis/modules/UNet',
    'hai/modules/loader/images_loader',
    'hai/modules/exporter/images_exporter',
]
external_folders = [
    # '~/VSProjects/particle_transformer',
    # '~/VSProjects/FINet',
    'repos']

# if config.API_FOLD_NAME in os.listdir('.'):
if CONST.API_FOLD_NAME in os.listdir('.'):
    external_folders.insert(0, f'{os.getcwd()}')

init_register(internal_modules=internal_modules, external_folders=external_folders)
# """






import glob
import os
import random
import shutil
import time
from pathlib import Path
from threading import Thread

try:
    import cv2
except:
    pass
import math
import numpy as np
from PIL import Image, ExifTags

from tqdm import tqdm

import damei as dm

# from damei.general import xyxy2xywh, xywh2xyxy, torch_distributed_zero_first

logger = dm.getLogger('dataset.py')

help_url = 'https://github.com/ultralytics/yolov5/wiki/Train-Custom-Data'
img_formats = ['.bmp', '.jpg', '.jpeg', '.png', '.tif', '.tiff', '.dng']
vid_formats = ['.mov', '.avi', '.mp4', '.mpg', '.mpeg', '.m4v', '.wmv', '.mkv']

# Get orientation exif tag
for orientation in ExifTags.TAGS.keys():
    if ExifTags.TAGS[orientation] == 'Orientation':
        break


def get_hash(files):
    # Returns a single hash value of a list of files
    return sum(os.path.getsize(f) for f in files if os.path.isfile(f))


def exif_size(img):
    # Returns exif-corrected PIL size
    s = img.size  # (width, height)
    try:
        rotation = dict(img._getexif().items())[orientation]
        if rotation == 6:  # rotation 270
            s = (s[1], s[0])
        elif rotation == 8:  # rotation 90
            s = (s[1], s[0])
    except:
        pass

    return s


def create_dataloader(path, imgsz, batch_size, stride, opt, hyp=None, augment=False, cache=False, pad=0.0, rect=False,
                      rank=-1, world_size=1, workers=8):
    # Make sure only the first process in DDP process the dataset first, and the following others can use the cache.
    from .dataset_torch import LoadImagesAndLabels, InfiniteDataLoader
    import torch

    with dm.general.torch_distributed_zero_first(rank):
        dataset = LoadImagesAndLabels(path, imgsz, batch_size,
                                      augment=augment,  # augment images
                                      hyp=hyp,  # augmentation hyperparameters
                                      rect=rect,  # rectangular training
                                      cache_images=cache,
                                      single_cls=opt.single_cls,
                                      stride=int(stride),
                                      pad=pad,
                                      rank=rank)

    batch_size = min(batch_size, len(dataset))
    nw = min([os.cpu_count() // world_size, batch_size if batch_size > 1 else 0, workers])  # number of workers
    sampler = torch.utils.data.distributed.DistributedSampler(dataset) if rank != -1 else None
    dataloader = InfiniteDataLoader(dataset,
                                    batch_size=batch_size,
                                    num_workers=nw,
                                    sampler=sampler,
                                    pin_memory=True,
                                    collate_fn=LoadImagesAndLabels.collate_fn)  # torch.utils.data.DataLoader()
    return dataloader, dataset

class _RepeatSampler(object):
    """ Sampler that repeats forever.

    Args:
        sampler (Sampler)
    """

    def __init__(self, sampler):
        self.sampler = sampler

    def __iter__(self):
        while True:
            yield from iter(self.sampler)


class LoadImages:  # for inference
    def __init__(self, path, img_size=640, video_sampling=5):
        self.count = 0
        p = str(Path(path))  # os-agnostic
        p = os.path.abspath(p)  # absolute path
        if '*' in p:
            files = sorted(glob.glob(p, recursive=True))  # glob
        elif os.path.isdir(p):
            files = sorted(glob.glob(os.path.join(p, '*.*')))  # dir
        elif os.path.isfile(p):
            files = [p]  # files
        else:
            raise Exception('ERROR: %s does not exist' % p)

        images = [x for x in files if os.path.splitext(x)[-1].lower() in img_formats]
        videos = [x for x in files if os.path.splitext(x)[-1].lower() in vid_formats]
        ni, nv = len(images), len(videos)

        self.img_size = img_size
        self.files = images + videos
        self.nf = ni + nv  # number of files
        self.video_flag = [False] * ni + [True] * nv
        self.mode = 'images'
        self.video_sampling = video_sampling
        if any(videos):
            self.new_video(videos[0])  # new video
        else:
            self.cap = None
        if not self.nf > 0:
            logger.warning('No images or videos found in %s', path)
        # assert self.nf > 0, 'No images or videos found in %s. Supported formats are:\nimages: %s\nvideos: %s' % \
        #                     (p, img_formats, vid_formats)

    def __iter__(self):
        self.count = 0
        return self

    def __next__(self):
        if self.count == self.nf:
            raise StopIteration
        path = self.files[self.count]

        if self.video_flag[self.count]:
            # Read video
            self.mode = 'video'

            while True:
                ret_val, img0 = self.cap.read()
                if not ret_val:  # 文件末尾，停止
                    self.count += 1
                    self.cap.release()
                    if self.count == self.nf:  # last video
                        raise StopIteration
                    else:
                        path = self.files[self.count]
                        self.new_video(path)
                        ret_val, img0 = self.cap.read()
                else:
                    if self.frame % self.video_sampling == 0:
                        self.frame += 1
                        break
                    else:
                        self.frame += 1

            print('video %g/%g (%g/%g) %s: ' % (self.count + 1, self.nf, self.frame, self.nframes, path), end='')

        else:
            # Read image
            self.count += 1
            img0 = cv2.imread(path)  # BGR
            assert img0 is not None, 'Image Not Found ' + path
            # print('image %g/%g %s: ' % (self.count, self.nf, path), end='')

        # Padded resize
        img = letterbox(img0, new_shape=self.img_size)[0]

        # Convert
        img = img[:, :, ::-1].transpose(2, 0, 1)  # BGR to RGB, to 3x416x416
        img = np.ascontiguousarray(img)

        # cv2.imwrite(path + '.letterbox.jpg', 255 * img.transpose((1, 2, 0))[:, :, ::-1])  # save letterbox image
        return path, img, img0, self.cap

    def new_video(self, path):
        self.frame = 0
        self.cap = cv2.VideoCapture(path)
        self.nframes = int(self.cap.get(cv2.CAP_PROP_FRAME_COUNT))

    def __len__(self):
        return self.nf  # number of files


class LoadWebcam:  # for inference
    def __init__(self, pipe=0, img_size=640):
        self.img_size = img_size

        if pipe == '0':
            pipe = 0  # local camera
        # pipe = 'rtsp://192.168.1.64/1'  # IP camera
        # pipe = 'rtsp://username:password@192.168.1.64/1'  # IP camera with login
        # pipe = 'rtsp://170.93.143.139/rtplive/470011e600ef003a004ee33696235daa'  # IP traffic camera
        # pipe = 'http://wmccpinetop.axiscam.net/mjpg/video.mjpg'  # IP golf camera

        # https://answers.opencv.org/question/215996/changing-gstreamer-pipeline-to-opencv-in-pythonsolved/
        # pipe = '"rtspsrc location="rtsp://username:password@192.168.1.64/1" latency=10 ! appsink'  # GStreamer

        # https://answers.opencv.org/question/200787/video-acceleration-gstremer-pipeline-in-videocapture/
        # https://stackoverflow.com/questions/54095699/install-gstreamer-support-for-opencv-python-package  # install help
        # pipe = "rtspsrc location=rtsp://root:root@192.168.0.91:554/axis-media/media.amp?videocodec=h264&resolution=3840x2160 protocols=GST_RTSP_LOWER_TRANS_TCP ! rtph264depay ! queue ! vaapih264dec ! videoconvert ! appsink"  # GStreamer

        self.pipe = pipe
        self.cap = cv2.VideoCapture(pipe)  # video capture object
        self.cap.set(cv2.CAP_PROP_BUFFERSIZE, 3)  # set buffer size

    def __iter__(self):
        self.count = -1
        return self

    def __next__(self):
        self.count += 1
        if cv2.waitKey(1) == ord('q'):  # q to quit
            self.cap.release()
            cv2.destroyAllWindows()
            raise StopIteration

        # Read frame
        if self.pipe == 0:  # local camera
            ret_val, img0 = self.cap.read()
            img0 = cv2.flip(img0, 1)  # flip left-right
        else:  # IP camera
            n = 0
            while True:
                n += 1
                self.cap.grab()
                if n % 30 == 0:  # skip frames
                    ret_val, img0 = self.cap.retrieve()
                    if ret_val:
                        break

        # Print
        assert ret_val, 'Camera Error %s' % self.pipe
        img_path = 'webcam.jpg'
        print('webcam %g: ' % self.count, end='')

        # Padded resize
        img = letterbox(img0, new_shape=self.img_size)[0]

        # Convert
        img = img[:, :, ::-1].transpose(2, 0, 1)  # BGR to RGB, to 3x416x416
        img = np.ascontiguousarray(img)

        return img_path, img, img0, None

    def __len__(self):
        return 0


class LoadStreams:  # multiple IP or RTSP cameras
    def __init__(self, sources='streams.txt', img_size=640, flip=False):
        self.mode = 'images'
        self.img_size = img_size
        self.flip = flip

        if os.path.isfile(sources):
            with open(sources, 'r') as f:
                sources = [x.strip() for x in f.read().splitlines() if len(x.strip())]
        else:
            sources = [sources]

        n = len(sources)
        self.imgs = [None] * n
        self.sources = sources
        for i, s in enumerate(sources):
            # Start the thread to read frames from the video stream
            print('%g/%g: %s... ' % (i + 1, n, s), end='')
            cap = cv2.VideoCapture(eval(s) if s.isnumeric() else s)
            assert cap.isOpened(), 'Failed to open %s' % s
            w = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))
            h = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
            fps = cap.get(cv2.CAP_PROP_FPS) % 100
            while True:
                is_get, self.imgs[i] = cap.read()  # guarantee first frame
                # print(f'\n{is_get} {self.imgs[i]}\n')
                if is_get:
                    break

            thread = Thread(target=self.update, args=([i, cap]), daemon=True)
            print(' success (%gx%g at %.2f FPS).' % (w, h, fps))
            thread.start()
        print('')  # newline

        # check for common shapes
        s = np.stack([letterbox(x, new_shape=self.img_size)[0].shape for x in self.imgs], 0)  # inference shapes
        self.rect = np.unique(s, axis=0).shape[0] == 1  # rect inference if all shapes equal
        if not self.rect:
            print('WARNING: Different stream shapes detected. For optimal performance supply similarly-shaped streams.')

    def update(self, index, cap):
        # Read next stream frame in a daemon thread
        n = 0
        while cap.isOpened():
            n += 1
            # _, self.imgs[index] = cap.read()
            cap.grab()
            if n == 4:  # read every 4th frame
                if self.flip:
                    _, frame = cap.retrieve()
                    self.imgs[index] = cv2.rotate(frame, cv2.ROTATE_180)
                else:
                    _, self.imgs[index] = cap.retrieve()
                n = 0
            time.sleep(0.01)  # wait time

    def __iter__(self):
        self.count = -1
        return self

    def __next__(self):
        self.count += 1
        img0 = self.imgs.copy()  # [8, 1920, 1280, 3]
        if cv2.waitKey(1) == ord('q'):  # q to quit
            cv2.destroyAllWindows()
            raise StopIteration

        # Letterbox
        img = [letterbox(x, new_shape=self.img_size, auto=self.rect)[0] for x in img0]

        # Stack
        img = np.stack(img, 0)

        # Convert
        img = img[:, :, :, ::-1].transpose(0, 3, 1, 2)  # BGR to RGB, to bsx3x416x416
        img = np.ascontiguousarray(img)

        return self.sources, img, img0, None

    def __len__(self):
        return 0  # 1E12 frames = 32 streams at 30 FPS for 30 years

# Ancillary functions --------------------------------------------------------------------------------------------------
def load_image(self, index):
    # loads 1 image from dataset, returns img, original hw, resized hw
    img = self.imgs[index]
    if img is None:  # not cached
        path = self.img_files[index]
        img = cv2.imread(path)  # BGR
        assert img is not None, 'Image Not Found ' + path
        h0, w0 = img.shape[:2]  # orig hw
        r = self.img_size / max(h0, w0)  # resize image to img_size
        if r != 1:  # always resize down, only resize up if training with augmentation
            interp = cv2.INTER_AREA if r < 1 and not self.augment else cv2.INTER_LINEAR
            img = cv2.resize(img, (int(w0 * r), int(h0 * r)), interpolation=interp)
        return img, (h0, w0), img.shape[:2]  # img, hw_original, hw_resized
    else:
        return self.imgs[index], self.img_hw0[index], self.img_hw[index]  # img, hw_original, hw_resized


def augment_hsv(img, hgain=0.5, sgain=0.5, vgain=0.5):
    r = np.random.uniform(-1, 1, 3) * [hgain, sgain, vgain] + 1  # random gains
    hue, sat, val = cv2.split(cv2.cvtColor(img, cv2.COLOR_BGR2HSV))
    dtype = img.dtype  # uint8

    x = np.arange(0, 256, dtype=np.int16)
    lut_hue = ((x * r[0]) % 180).astype(dtype)
    lut_sat = np.clip(x * r[1], 0, 255).astype(dtype)
    lut_val = np.clip(x * r[2], 0, 255).astype(dtype)

    img_hsv = cv2.merge((cv2.LUT(hue, lut_hue), cv2.LUT(sat, lut_sat), cv2.LUT(val, lut_val))).astype(dtype)
    cv2.cvtColor(img_hsv, cv2.COLOR_HSV2BGR, dst=img)  # no return needed

    # Histogram equalization
    # if random.random() < 0.2:
    #     for i in range(3):
    #         img[:, :, i] = cv2.equalizeHist(img[:, :, i])


def load_mosaic(self, index):
    # loads images in a mosaic

    labels4 = []
    s = self.img_size
    yc, xc = [int(random.uniform(-x, 2 * s + x)) for x in self.mosaic_border]  # mosaic center x, y
    indices = [index] + [random.randint(0, len(self.labels) - 1) for _ in range(3)]  # 3 additional image indices
    for i, index in enumerate(indices):
        # Load image
        img, _, (h, w) = load_image(self, index)

        # place img in img4
        if i == 0:  # top left
            img4 = np.full((s * 2, s * 2, img.shape[2]), 114, dtype=np.uint8)  # base image with 4 tiles
            x1a, y1a, x2a, y2a = max(xc - w, 0), max(yc - h, 0), xc, yc  # xmin, ymin, xmax, ymax (large image)
            x1b, y1b, x2b, y2b = w - (x2a - x1a), h - (y2a - y1a), w, h  # xmin, ymin, xmax, ymax (small image)
        elif i == 1:  # top right
            x1a, y1a, x2a, y2a = xc, max(yc - h, 0), min(xc + w, s * 2), yc
            x1b, y1b, x2b, y2b = 0, h - (y2a - y1a), min(w, x2a - x1a), h
        elif i == 2:  # bottom left
            x1a, y1a, x2a, y2a = max(xc - w, 0), yc, xc, min(s * 2, yc + h)
            x1b, y1b, x2b, y2b = w - (x2a - x1a), 0, w, min(y2a - y1a, h)
        elif i == 3:  # bottom right
            x1a, y1a, x2a, y2a = xc, yc, min(xc + w, s * 2), min(s * 2, yc + h)
            x1b, y1b, x2b, y2b = 0, 0, min(w, x2a - x1a), min(y2a - y1a, h)

        img4[y1a:y2a, x1a:x2a] = img[y1b:y2b, x1b:x2b]  # img4[ymin:ymax, xmin:xmax]
        padw = x1a - x1b
        padh = y1a - y1b

        # Labels
        x = self.labels[index]
        labels = x.copy()
        if x.size > 0:  # Normalized xywh to pixel xyxy format
            labels[:, 1] = w * (x[:, 1] - x[:, 3] / 2) + padw
            labels[:, 2] = h * (x[:, 2] - x[:, 4] / 2) + padh
            labels[:, 3] = w * (x[:, 1] + x[:, 3] / 2) + padw
            labels[:, 4] = h * (x[:, 2] + x[:, 4] / 2) + padh
        labels4.append(labels)

    # Concat/clip labels
    if len(labels4):
        labels4 = np.concatenate(labels4, 0)
        np.clip(labels4[:, 1:], 0, 2 * s, out=labels4[:, 1:])  # use with random_perspective
        # img4, labels4 = replicate(img4, labels4)  # replicate

    # Augment
    img4, labels4 = random_perspective(img4, labels4,
                                       degrees=self.hyp['degrees'],
                                       translate=self.hyp['translate'],
                                       scale=self.hyp['scale'],
                                       shear=self.hyp['shear'],
                                       perspective=self.hyp['perspective'],
                                       border=self.mosaic_border)  # border to remove

    return img4, labels4


def replicate(img, labels):
    # Replicate labels
    h, w = img.shape[:2]
    boxes = labels[:, 1:].astype(int)
    x1, y1, x2, y2 = boxes.T
    s = ((x2 - x1) + (y2 - y1)) / 2  # side length (pixels)
    for i in s.argsort()[:round(s.size * 0.5)]:  # smallest indices
        x1b, y1b, x2b, y2b = boxes[i]
        bh, bw = y2b - y1b, x2b - x1b
        yc, xc = int(random.uniform(0, h - bh)), int(random.uniform(0, w - bw))  # offset x, y
        x1a, y1a, x2a, y2a = [xc, yc, xc + bw, yc + bh]
        img[y1a:y2a, x1a:x2a] = img[y1b:y2b, x1b:x2b]  # img4[ymin:ymax, xmin:xmax]
        labels = np.append(labels, [[labels[i, 0], x1a, y1a, x2a, y2a]], axis=0)

    return img, labels


def letterbox(img, new_shape=(640, 640), color=(114, 114, 114), auto=True, scaleFill=False, scaleup=True):
    # Resize image to a 32-pixel-multiple rectangle https://github.com/ultralytics/yolov3/issues/232
    if isinstance(new_shape, int):
        new_shape = (new_shape, new_shape)
    if img is None:
        logger.warn(f'Input img is None, padding in gray image.')
        # print(new_shape, type(new_shape))
        # print(new_shape[0])
        new_img = np.zeros((new_shape[0], new_shape[1], 3), dtype=np.uint8)
        new_img[...] = 114
        return new_img, (1, 1), (0, 0)
    shape = img.shape[:2]  # current shape [height, width]

    # Scale ratio (new / old)
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    if not scaleup:  # only scale down, do not scale up (for better test mAP)
        r = min(r, 1.0)

    # Compute padding
    ratio = r, r  # width, height ratios
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = new_shape[1] - new_unpad[0], new_shape[0] - new_unpad[1]  # wh padding
    if auto:  # minimum rectangle
        dw, dh = np.mod(dw, 32), np.mod(dh, 32)  # wh padding
    elif scaleFill:  # stretch
        dw, dh = 0.0, 0.0
        new_unpad = (new_shape[1], new_shape[0])
        ratio = new_shape[1] / shape[1], new_shape[0] / shape[0]  # width, height ratios

    dw /= 2  # divide padding into 2 sides
    dh /= 2

    if shape[::-1] != new_unpad:  # resize
        img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=color)  # add border
    # print(img.shape, ratio, (dw, dh))
    return img, ratio, (dw, dh)


def random_perspective(img, targets=(), degrees=10, translate=.1, scale=.1, shear=10, perspective=0.0, border=(0, 0)):
    # torchvision.transforms.RandomAffine(degrees=(-10, 10), translate=(.1, .1), scale=(.9, 1.1), shear=(-10, 10))
    # targets = [cls, xyxy]

    height = img.shape[0] + border[0] * 2  # shape(h,w,c)
    width = img.shape[1] + border[1] * 2

    # Center
    C = np.eye(3)
    C[0, 2] = -img.shape[1] / 2  # x translation (pixels)
    C[1, 2] = -img.shape[0] / 2  # y translation (pixels)

    # Perspective
    P = np.eye(3)
    P[2, 0] = random.uniform(-perspective, perspective)  # x perspective (about y)
    P[2, 1] = random.uniform(-perspective, perspective)  # y perspective (about x)

    # Rotation and Scale
    R = np.eye(3)
    a = random.uniform(-degrees, degrees)
    # a += random.choice([-180, -90, 0, 90])  # add 90deg rotations to small rotations
    s = random.uniform(1 - scale, 1 + scale)
    # s = 2 ** random.uniform(-scale, scale)
    R[:2] = cv2.getRotationMatrix2D(angle=a, center=(0, 0), scale=s)

    # Shear
    S = np.eye(3)
    S[0, 1] = math.tan(random.uniform(-shear, shear) * math.pi / 180)  # x shear (deg)
    S[1, 0] = math.tan(random.uniform(-shear, shear) * math.pi / 180)  # y shear (deg)

    # Translation
    T = np.eye(3)
    T[0, 2] = random.uniform(0.5 - translate, 0.5 + translate) * width  # x translation (pixels)
    T[1, 2] = random.uniform(0.5 - translate, 0.5 + translate) * height  # y translation (pixels)

    # Combined rotation matrix
    M = T @ S @ R @ P @ C  # order of operations (right to left) is IMPORTANT
    if (border[0] != 0) or (border[1] != 0) or (M != np.eye(3)).any():  # image changed
        if perspective:
            img = cv2.warpPerspective(img, M, dsize=(width, height), borderValue=(114, 114, 114))
        else:  # affine
            img = cv2.warpAffine(img, M[:2], dsize=(width, height), borderValue=(114, 114, 114))

    # Visualize
    # import matplotlib.pyplot as plt
    # ax = plt.subplots(1, 2, figsize=(12, 6))[1].ravel()
    # ax[0].imshow(img[:, :, ::-1])  # base
    # ax[1].imshow(img2[:, :, ::-1])  # warped

    # Transform label coordinates
    n = len(targets)
    if n:
        # warp points
        xy = np.ones((n * 4, 3))
        xy[:, :2] = targets[:, [1, 2, 3, 4, 1, 4, 3, 2]].reshape(n * 4, 2)  # x1y1, x2y2, x1y2, x2y1
        xy = xy @ M.T  # transform
        if perspective:
            xy = (xy[:, :2] / xy[:, 2:3]).reshape(n, 8)  # rescale
        else:  # affine
            xy = xy[:, :2].reshape(n, 8)

        # create new boxes
        x = xy[:, [0, 2, 4, 6]]
        y = xy[:, [1, 3, 5, 7]]
        xy = np.concatenate((x.min(1), y.min(1), x.max(1), y.max(1))).reshape(4, n).T

        # # apply angle-based reduction of bounding boxes
        # radians = a * math.pi / 180
        # reduction = max(abs(math.sin(radians)), abs(math.cos(radians))) ** 0.5
        # x = (xy[:, 2] + xy[:, 0]) / 2
        # y = (xy[:, 3] + xy[:, 1]) / 2
        # w = (xy[:, 2] - xy[:, 0]) * reduction
        # h = (xy[:, 3] - xy[:, 1]) * reduction
        # xy = np.concatenate((x - w / 2, y - h / 2, x + w / 2, y + h / 2)).reshape(4, n).T

        # clip boxes
        xy[:, [0, 2]] = xy[:, [0, 2]].clip(0, width)
        xy[:, [1, 3]] = xy[:, [1, 3]].clip(0, height)

        # filter candidates
        i = box_candidates(box1=targets[:, 1:5].T * s, box2=xy.T)
        targets = targets[i]
        targets[:, 1:5] = xy[i]

    return img, targets


def box_candidates(box1, box2, wh_thr=2, ar_thr=20, area_thr=0.1):  # box1(4,n), box2(4,n)
    # Compute candidate boxes: box1 before augment, box2 after augment, wh_thr (pixels), aspect_ratio_thr, area_ratio
    w1, h1 = box1[2] - box1[0], box1[3] - box1[1]
    w2, h2 = box2[2] - box2[0], box2[3] - box2[1]
    ar = np.maximum(w2 / (h2 + 1e-16), h2 / (w2 + 1e-16))  # aspect ratio
    return (w2 > wh_thr) & (h2 > wh_thr) & (w2 * h2 / (w1 * h1 + 1e-16) > area_thr) & (ar < ar_thr)  # candidates


def cutout(image, labels):
    # Applies image cutout augmentation https://arxiv.org/abs/1708.04552
    h, w = image.shape[:2]

    def bbox_ioa(box1, box2):
        # Returns the intersection over box2 area given box1, box2. box1 is 4, box2 is nx4. boxes are x1y1x2y2
        box2 = box2.transpose()

        # Get the coordinates of bounding boxes
        b1_x1, b1_y1, b1_x2, b1_y2 = box1[0], box1[1], box1[2], box1[3]
        b2_x1, b2_y1, b2_x2, b2_y2 = box2[0], box2[1], box2[2], box2[3]

        # Intersection area
        inter_area = (np.minimum(b1_x2, b2_x2) - np.maximum(b1_x1, b2_x1)).clip(0) * \
                     (np.minimum(b1_y2, b2_y2) - np.maximum(b1_y1, b2_y1)).clip(0)

        # box2 area
        box2_area = (b2_x2 - b2_x1) * (b2_y2 - b2_y1) + 1e-16

        # Intersection over box2 area
        return inter_area / box2_area

    # create random masks
    scales = [0.5] * 1 + [0.25] * 2 + [0.125] * 4 + [0.0625] * 8 + [0.03125] * 16  # image size fraction
    for s in scales:
        mask_h = random.randint(1, int(h * s))
        mask_w = random.randint(1, int(w * s))

        # box
        xmin = max(0, random.randint(0, w) - mask_w // 2)
        ymin = max(0, random.randint(0, h) - mask_h // 2)
        xmax = min(w, xmin + mask_w)
        ymax = min(h, ymin + mask_h)

        # apply random color mask
        image[ymin:ymax, xmin:xmax] = [random.randint(64, 191) for _ in range(3)]

        # return unobscured labels
        if len(labels) and s > 0.03:
            box = np.array([xmin, ymin, xmax, ymax], dtype=np.float32)
            ioa = bbox_ioa(box, labels[:, 1:5])  # intersection over area
            labels = labels[ioa < 0.60]  # remove >60% obscured labels

    return labels


def reduce_img_size(path='path/images', img_size=1024):  # from utils.datasets import *; reduce_img_size()
    # creates a new ./images_reduced folder with reduced size images of maximum size img_size
    path_new = path + '_reduced'  # reduced images path
    create_folder(path_new)
    for f in tqdm(glob.glob('%s/*.*' % path)):
        try:
            img = cv2.imread(f)
            h, w = img.shape[:2]
            r = img_size / max(h, w)  # size ratio
            if r < 1.0:
                img = cv2.resize(img, (int(w * r), int(h * r)), interpolation=cv2.INTER_AREA)  # _LINEAR fastest
            fnew = f.replace(path, path_new)  # .replace(Path(f).suffix, '.jpg')
            cv2.imwrite(fnew, img)
        except:
            print('WARNING: image failure %s' % f)


def recursive_dataset2bmp(dataset='path/dataset_bmp'):  # from utils.datasets import *; recursive_dataset2bmp()
    # Converts dataset to bmp (for faster training)
    formats = [x.lower() for x in img_formats] + [x.upper() for x in img_formats]
    for a, b, files in os.walk(dataset):
        for file in tqdm(files, desc=a):
            p = a + '/' + file
            s = Path(file).suffix
            if s == '.txt':  # replace text
                with open(p, 'r') as f:
                    lines = f.read()
                for f in formats:
                    lines = lines.replace(f, '.bmp')
                with open(p, 'w') as f:
                    f.write(lines)
            elif s in formats:  # replace image
                cv2.imwrite(p.replace(s, '.bmp'), cv2.imread(p))
                if s != '.bmp':
                    os.system("rm '%s'" % p)


def imagelist2folder(path='path/images.txt'):  # from utils.datasets import *; imagelist2folder()
    # Copies all the images in a text file (list of images) into a folder
    create_folder(path[:-4])
    with open(path, 'r') as f:
        for line in f.read().splitlines():
            os.system('cp "%s" %s' % (line, path[:-4]))
            print(line)


def create_folder(path='./new'):
    # Create folder
    if os.path.exists(path):
        shutil.rmtree(path)  # delete output folder
    os.makedirs(path)  # make new output folder




import os, sys
import hai
# from hai import Testor

# module_path = '/home/zzd/VSProjects/FINet'

# cwd = os.getcwd()

# os.chdir(module_path)
# if module_path not in sys.path:
#     sys.path.insert(0, module_path)
# testor = Testor()
# testor.test_module(module_name='FINet')






import argparse
import damei as dm


def get_args():
    parser = dm.argparse.ArgumentParser(description='Predict masks from input images')
    
    # train args
    parser.add_argument('-s', '--source', type=str, 
                            default='~/datasets/hai_datasets/carvana', 
                            help='source path of images and masks')
    parser.add_argument('--epochs', '-e', metavar='E', type=int, default=5, help='Number of epochs')
    parser.add_argument('--batch-size', '-b', dest='batch_size', metavar='B', type=int, default=4, help='Batch size')
    parser.add_argument('--learning-rate', '-l', metavar='LR', type=float, default=1e-5,
                        help='Learning rate', dest='lr')
    parser.add_argument('--weights', '-w', type=str, default=None, help='trained model weights path')
    parser.add_argument('--scale', type=float, default=0.5, help='Downscaling factor of the images')
    parser.add_argument('--validation', '-v', dest='val', type=float, default=10.0,
                        help='Percent of the data that is used as validation (0-100)')
    parser.add_argument('--amp', action='store_true', default=False, help='Use mixed precision')
    parser.add_argument('--bilinear', action='store_true', default=False, help='Use bilinear upsampling')
    parser.add_argument('--classes', '-c', type=int, default=2, help='Number of classes')
    parser.add_argument('--lr', type=float, default=1e-5, help='learning rate')
    # parser.add_argument('--val', type=float, default=10.0, help='validation percent')
    # infer args

    # parser.add_argument('--model', '-m', default='MODEL.pth', metavar='FILE',
    #                     help='Specify the file in which the model is stored')
    # parser.add_argument('--input', '-i', metavar='INPUT', nargs='+', help='Filenames of input images', required=True)
    parser.add_argument('--output', '-o', default='runs/outputs', help='Filenames of output images')
    parser.add_argument('--viz', action='store_true',
                        help='Visualize the images as they are processed')
    parser.add_argument('--no-save', '-n', action='store_true', help='Do not save the output masks')
    parser.add_argument('--mask-threshold', '-t', type=float, default=0.5,
                        help='Minimum probability value to consider a mask pixel white')
    
    parser.add_argument('--device', '-d', default='cuda', help='Device to use for inference')

    return parser.parse_args()


"""
测试
"""

# ! /usr/bin/env python
# coding=utf8
import os, sys
from pathlib import Path
import grpc
from grpc import ssl_server_credentials
import time
import damei as dm
from concurrent import futures

pydir = Path(os.path.abspath(__file__)).parent
sys.path.append(f'{pydir}')
import xai_pb2_grpc, xai_pb2
sys.path.remove(f'{pydir}')
import xsensing_ai as xai


logger = dm.getLogger('xai_server')
_ONE_DAY_IN_SECONDS = 60 * 60 * 24


class XAIService(xai_pb2_grpc.GrpcServiceServicer):
    '''
    继承GrpcServiceServicer,实现hello方法
    '''

    def __init__(self):
        self.uaii = xai.UAII()

    def hello(self, request, context):
        '''
        具体实现hello的方法，并按照pb的返回对象构造HelloResponse返回
        :param request:
        :param context: 是保留字段，不用管
        :return:
        '''
        # logger.info(f'hello request: {request}')
        logger.info(f'grpc func "hello" is called.')
        result = request.data + request.skill.name + " this is gprc test service"
        list_result = {"12": 1232}
        return xai_pb2.HelloResponse(result=str(result),
                                     map_result=list_result)

    def call(self, request, context):
        func = request.func

        # 把google.protobuf.xx。ScalarMapContainer转换成dict
        # 弃用，原因是该方法支持map传参，但map传参不支持多种数据类型，必须指定一种类型
        # params = dict(request.params)

        # 处理bytes传参，传是的一个Json字典
        params = request.params
        try:
            params = eval(params.decode('utf-8'))
        except:
            params = params.decode('utf-8')
        params = dict(params)
        # params = dict(eval(params.decode('utf-8')))  # 解析后是一个dict
        # print(params, type(params))

        logger.info(f'grpc func "call" is called. func_name: "{func}", params: {params}')
        status, data = self._call(func, params)
        data = str(data).encode('utf-8')
        return xai_pb2.CallResponse(status=status, data=data)

    def _call(self, func, params):
        """
        实际调用的方法
        :param func: 函数名
        :param params: 参数
        :return: status 和 data
        """
        try:
            if func == 'ps':
                res = self.uaii.ps(**params)
                s = 1
                data = res
            elif func == "build_stream":
                print(params)
                s, data = self.uaii.build_stream(cfg=params, is_req=True)  # 返回的res是stream对象
            elif func == "get_stream_info":
                stream_name = params.pop('stream_name', params['stream_name'])  # 必选参数
                s, data = self.uaii.get_stream_info(stream=stream_name, is_req=True, **params)
            elif func == "get_stream_cfg":
                stream_name = params.pop('stream_name', params['stream_name'])  # 必选参数
                addr = params.pop('addr', '/')  # 可选参数
                s, data = self.uaii.get_stream_cfg(stream=stream_name, addr=addr, is_req=True, **params)
            elif func == 'set_stream_cfg':
                stream_name = params.pop('stream_name', params['stream_name'])  # 必选参数
                addr = params.pop('addr', '/')  # 可选参数
                cfg = params.pop('cfg', params['cfg'])  # 必选参数
                s, data = self.uaii.set_stream_cfg(stream=stream_name, addr=addr, cfg=cfg, is_req=True, **params)
            elif func == "start_stream":
                stream_name = params.pop('stream_name', params['stream_name'])
                s, data = self.uaii.start_stream(stream=stream_name, is_req=True, **params)  # 返回的res是stream对象
            else:
                raise Exception(f'Call func: "{func}" is not supported.')
            return s, data
        except KeyError as e:
            logger.error(f'Call func: "{func}" params failed. KeyError: {e}')
            return -1, f'Call func: "{func}" params failed. KeyError: {e}'
        except NotImplementedError as e:
            logger.error(f'Call func: "{func}" params failed. NotImplementedError: {e}')
            return -1, f'Call func: "{func}" params failed. NotImplementedError: {e}'
        except Exception as e:
            logger.error(f'Call func: "{func}" failed. Exception: {e}')
            return -1, f'Call func: "{func}" failed. Exception: {e}'


def run(ip='localhost', port=50052):
    '''
    模拟服务启动
    :return:
    '''
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    xai_pb2_grpc.add_GrpcServiceServicer_to_server(XAIService(), server)
    # address = f'localhost:{port}'
    address = f'{ip}:{port}'
    # server.add_insecure_port(f'[::]:{port}')
    server.add_insecure_port(address)
    server.start()
    # print("start service...")
    logger.info(f"Service is started at {address} ...")
    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        server.stop(0)


def run_secure(port=50052):
    """开启安全的服务端"""
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
    xai_pb2_grpc.add_GrpcServiceServicer_to_server(XAIService(), server)



def run_debug():
    from xsensing_ai.uaii.server.grpc.grpc_xai_client import XAIGrpcClient
    client = XAIGrpcClient('localhost', 50052)
    # response = client.hello()
    # print("received:", response.result, response.map_result)
    response = client.__call__()
    print(f'call的返回结果：{response.result}')


if __name__ == '__main__':
    run(port=10086)
